__author__ = 'quentin'

from idoc.server.core.base import DescribedObject
from idoc.server.core.variables import BaseIntVariable
from idoc.server.hardware.interfaces.interfaces import DefaultInterface
from idoc.server.utils.scheduler import Scheduler



class HasInteractedVariable(BaseIntVariable):
    """
    Custom variable to save whether the stimulator has sent instruction to its hardware interface. 0 means
     no interaction. Any positive integer describes a different interaction.
    """
    functional_type = "interaction"
    header_name = "has_interacted"



class BaseStimulator(DescribedObject):
    _tracker = None
    _HardwareInterfaceClass = None

    def __init__(self, hardware_connection, date_range=""):
        """
        Template class to interact with the tracked animal in a real-time feedback loop.
        Derived classes must have an attribute ``_hardwareInterfaceClass`` defining the class of the
        :class:`~ethoscope.hardware.interfaces.interfaces.BaseInterface` object (not on object) that instances will
        share with one another. In addition, they must implement a ``_decide()`` method.

        :param hardware_connection: The hardware interface to use.
        :type hardware_connection: :class:`~ethoscope.hardware.interfaces.interfaces.BaseInterface`
        :param date_range: the start and stop date/time for the stimulator. Format described `here <https://github.com/gilestrolab/ethoscope/blob/master/user_manual/schedulers.md>`_
        :type date_range: str

        """

        self._scheduler = Scheduler(date_range)
        self._hardware_connection = hardware_connection

    def apply(self):
        """
        Apply this stimulator. This method will:

        1. check ``_tracker`` exists
        2. decide (``_decide``) whether to interact
        3. if 2. pass the interaction arguments to the hardware interface

        :return: whether a stimulator has worked, and a result dictionary
        """
        if self._tracker is None:
            raise ValueError("No tracker bound to this stimulator. Use `bind_tracker()` methods")

        if self._scheduler.check_time_range() is False:
            return HasInteractedVariable(False) , {}
        interact, result  = self._decide()
        if interact > 0:
            self._deliver(**result)

        return interact, result


    def bind_tracker(self, tracker):
        """
        Link a tracker to this interactor

        :param tracker: a tracker object.
        :type tracker: :class:`~ethoscope.trackers.trackers.BaseTracker`
        """
        self._tracker = tracker

    def _decide(self):
        raise NotImplementedError

    def _deliver(self, **kwargs):
        if self._hardware_connection is not None:
            self._hardware_connection.send_instruction(kwargs)


class DefaultStimulator(BaseStimulator):
    """
    Default interactor. Simply never interacts
    """
    _description = {"overview": "The default 'interactor'. To use when no hardware interface is to be used.",
                    "arguments": []}
    _HardwareInterfaceClass = DefaultInterface

    def _decide(self):
        out = HasInteractedVariable(False)
        return out, {}





# Standard library imports
import warnings
import logging
import os.path
import traceback

import pandas as pd

# Local application imports
from idoc.server.controllers.controllers import Controller
from idoc.server.core.base import Base, Status
from idoc.server.core.recognizer import Recognizer
# from idoc.server.trackers.simple_tracker import SimpleTracker
from idoc.server.trackers.adaptive_bg_tracker import AdaptiveBGModel
from idoc.server.io.result_writer import CSVResultWriter
from idoc.configuration import IDOCConfiguration
from idoc.helpers import get_machine_id, get_machine_name
from idoc.server.io.pylon_camera import PylonCamera
from idoc.server.io.opencv_camera import OpenCVCamera
from idoc.server.roi_builders.roi_builders import DefaultROIBuilder
from idoc.server.roi_builders.high_contrast_roi_builder import HighContrastROIBuilder
from idoc.server.controllers.boards import ArduinoDummy, ArduinoMegaBoard, ArduinoBoard
from idoc.server.roi_builders.target_roi_builder import IDOCROIBuilder
from idoc.server.drawers.drawers import DefaultDrawer
from idoc.debug import IDOCException
from idoc.helpers import hours_minutes_seconds, iso_format
config = IDOCConfiguration()
logger = logging.getLogger(__name__)

from idoc.configuration import IDOCConfiguration
config = IDOCConfiguration()

class ControlThread(Base):
    r"""
    A class to orchestrate the communication between the server
    and all the IDOC modules
    """

    _option_dict = {

        'board': {
            'possible_classes': [
                ArduinoBoard, ArduinoMegaBoard, ArduinoDummy
            ]
        },

        'camera': {
            'possible_classes': [
                PylonCamera, OpenCVCamera
            ]
        },

        'drawer': {
            'possible_classes': [
                DefaultDrawer
            ]
        },

        'result_writer': {
            'possible_classes': [
                CSVResultWriter
            ]
        },

        'roi_builder': {
            'possible_classes': [
                IDOCROIBuilder, HighContrastROIBuilder, DefaultROIBuilder
            ]
        },

        'tracker': {
            'possible_classes': [
                AdaptiveBGModel
            ]
        },

    }

    sampling_rate = 50 # overriden by conf


    valid_actions = ["start", "stop", "warm_up", "start_experiment", "stop_experiment", "reset_experiment"]

    def __init__(self, machine_id, version, result_dir, user_data, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._settings = {}

        if user_data["adaptation_time"] is None:
            adaptation_time = config.content["experiment"]["adaptation_time"]
        else:
            adaptation_time = user_data["adaptation_time"]

        self._settings.update({
            "adaptation_time": adaptation_time
        })

        self._submodules = {"recognizer": None, "controller": None, "result_writer": None}

        self.machine_id = machine_id
        self.machine_name = get_machine_name()

        self.version = version
        self._result_dir = result_dir

        self._location = config.content['experiment']['location']
        self._max_duration = user_data['max_duration'] or config.content['experiment']['max_duration']

        ## Assignment of attributes
        self.control = user_data['control']
        self.recognize = user_data['recognize']

        # self.reporting = user_data["reporting"]

        self._user_classes = {
            "camera": user_data.pop("camera"),
            "board": user_data.pop("board"),
            "drawer": user_data.pop("drawer"),
            "result_writer": user_data.pop("result_writer"),
            "roi_builder": user_data.pop("roi_builder"),
            "tracker": user_data.pop("tracker")
        }
        self._video_out = None
        self._user_data = user_data
        self._status = {}
        self.description = ""
        self._warmed_up = False
        logger.info("Control thread initialized")
        self.initialise_submodules()
        logger.info("Control thread submodules initialized")

    @property
    def selected_options(self):
        return {
            "camera": self._pick_class("camera"),
            "board": self._pick_class("board"),
            "drawer": self._pick_class("drawer"),
            "result_writer": self._pick_class("result_writer"),
            "roi_builder": self._pick_class("roi_builder"),
            "tracker": self._pick_class("tracker"),
            "controller": self.controlling,
            "recognizing": self.recognizing
        }

    @property
    def run_id(self):
        return "%s_%s" % (self._start_datetime, self.machine_id)

    @property
    def run_id_short(self):
        return "%s_%s" % (self._start_datetime, self.machine_id[:5])

    @property
    def adaptation_time(self):
        return self._settings['adaptation_time']

    @adaptation_time.setter
    def adaptation_time(self, adaptation_time):
        self._settings['adaptation_time'] = adaptation_time * 60

    @property
    def adaptation_time_human(self):
        return iso_format(
            hours_minutes_seconds(
                self._settings['adaptation_time']
            )
        )

    @property
    def adaptation_offset(self):
        return max(0, self.adaptation_time - self.elapsed_seconds)

    @property
    def metadata(self):
        metadata = pd.DataFrame([
            {"field": "run_id", "value": self.run_id_short},
            {"field": "machine_id", "value": self.machine_id},
            {"field": "machine_name", "value": self.machine_name},
            {"field": "version", "value": self.version},
            {"field": "date_time", "value": self.start_datetime},
            {"field": "selected_options", "value": self.selected_options},
            {"field": "settings", "value": self.settings},
            {"field": "user_data", "value": self._user_data},
            {"field": "description", "value": self.description}
        ])

        return metadata

    @property
    def info(self):
        return self._info

    @info.getter
    def info(self):

        if self.recognize:
            self._info["recognizer"] = self.recognizer.info
        else:
            self._info["recognizer"] = {"status": "offline"}

        if self.control:
            self._info["controller"] = self.controller.info
        else:
            self._info["controller"] = {"status": "offline"}

        self._info["result_writer"] = self.result_writer.info

        self._info.update({
            "id": self.machine_id,
            "location": self._location,
            "name": self.machine_name,
            "status": self.status["control_thread"],
            "clock": self.clock,
            "settings": self.settings,
            "result_writer": {"status": self.result_writer.status}

        })

        return self._info

    @property
    def controlling(self):
        if self.controller is not None  and self.controller is not None:
            return self.controller.status == 'running'
        return False

    @property
    def recognizing(self):
        if self.recognize is not None and self.recognizer is not None:
            return self.recognizer.status == 'running'
        return False

    @property
    def status(self):
        return self._status

    @status.getter
    def status(self):

        if not self.running:
            self._status["control_thread"] = "idle"

        elif self.ready and not self.running:
            self._status["control_thread"] = 'ready'

        elif self.running and not self.stopped:
            self._status["control_thread"] = "running"

        elif self.stopped:
            self._status["control_thread"] = "stopped"

        else:
            logger.warning('%s status undefined', self.__class__.__name__)
            self._status["control_thread"] = 'undefined'

        return self._status


    def load_paradigm(self, paradigm_path):
        self.controller.load_paradigm(paradigm_path)

    def list_paradigms(self):
        # TODO remove me
        logger.info("Running controller.list()")
        return self.controller.list()

    def toggle(self, hardware, value):

        result = {"status": None}

        if self.controller is not None:
            try:
                self.controller.toggle(hardware, value)
                return {"status": "success"}
            except Exception as error:
                logger.warning(error)
                logger.warning(traceback.print_exc())
        return result

    @property
    def video_out(self):
        filename = self.run_id + ".avi"
        self._video_out = os.path.join(
            self.result_writer.result_dir,
            filename
        )

        return self._video_out

    @property
    def mapping(self):
        if self.control:
            return self.controller.mapping
        return None

    @property
    def pin_state(self):
        if self.control:
            return self.controller.pin_state

    @property
    def recognizer(self):

        return self._submodules["recognizer"]

    @recognizer.setter
    def recognizer(self, recognizer):
        self._submodules["recognizer"] = recognizer
        self._settings['recognizer'] = recognizer.settings


    @property
    def controller(self):
        if self.control:
            return self._submodules["controller"]

    @controller.setter
    def controller(self, controller):
        self._submodules["controller"] = controller
        self._settings['controller'] = controller.settings

    @property
    def result_writer(self):
        return self._submodules["result_writer"]

    @result_writer.setter
    def result_writer(self, result_writer):
        self._submodules["result_writer"] = result_writer

    def command(self, submodule, action):
        r"""
        Control the submodules with actions
        """
        try:
            if submodule == 'control_thread':
                instance = self

            elif submodule in self._submodules:
                instance = self._submodules[submodule]
            else:
                logger.warning(
                    "Please command one of the following modules: %s",
                    " ".join(
                        ["control_thread"] + list(self._submodules.keys())
                    )
                )

            if action == "start" and instance.running:
                logger.warning("%s is already running", instance.__class__.__name__)
                return self.status

            else:
                logger.info("Running action %s on module %s", action, instance)
                getattr(instance, action)()

        except RuntimeError as error:
            logger.warning("You passed action: %s. %s is already started", action, submodule)

        except Exception as error:
            logger.warning(error)
            logger.warning(traceback.print_exc())
            logger.info("Action %s is not valid. Please select one in %s.", action, ' '.join(instance.valid_actions))


        return self.status

    def _pick_class(self, category):
        """
        Select a class for a given category
        among those listed in the _option_dict default_class key
        for that category.
        """

        class_name = self._user_classes[category] or config.content['default_class'][category]

        for cls in self._option_dict[category]['possible_classes']:
            if class_name == cls.__name__:
                return cls

        raise IDOCException("Class %s is not a possible class for category %s" % (class_name, category))

    def initialise_result_writer(self):
        logger.info('Starting result writer')
        result_writer_class = self._pick_class("result_writer")
        result_writer_args = config.content['io']['result_writer']['args']
        result_writer_kwargs = config.content['io']['result_writer']['kwargs']

        self.result_writer = result_writer_class(
            *result_writer_args,
            **result_writer_kwargs
        )

        if self._user_data['start_datetime'] is not None:
            self.result_writer.start_datetime = self._user_data['start_datetime']


    def initialise_controller(self):
        logger.debug('Starting controller')

        self.controller = Controller(
            mapping_path=config.content["controller"]["mapping_path"],
            paradigm_path=config.content["controller"]["paradigm_path"],
            result_writer=self.result_writer,
            board_class=self._pick_class("board"),
            arduino_port=config.content["controller"]["arduino_port"],
            use_wall_clock=self._user_data["use_wall_clock"]
        )
        # self.controller.experiment_start = self.experiment_start
        self._settings['controller'] = self.controller.settings


    def initialise_recognizer(self):
        logger.debug('Starting recognizer function')

        # Initialize camera (use the class declared in the config)
        camera_class = self._pick_class("camera")

        # Initialize roi_builder
        roi_builder_class = self._pick_class("roi_builder")
        roi_builder_args = config.content['roi_builder']['args']
        roi_builder_kwargs = config.content['roi_builder']['kwargs']
        roi_builder = roi_builder_class(
            *roi_builder_args,
            **roi_builder_kwargs
        )

        # Initialize drawer
        drawer_class = self._pick_class("drawer")
        drawer_kwargs = config.content['drawer']['kwargs']


        if self._user_data['start_datetime'] is not None:
            start_datetime = self._user_data['start_datetime']
            drawer_kwargs['video_out'] = os.path.join(self._result_dir, self.machine_id, self.machine_name, start_datetime, '%s_%s.avi' % (start_datetime, self.machine_id))


        # Pick tracker class and initialize recognizer
        tracker_class = self._pick_class("tracker")
        recognizer = Recognizer(
            tracker_class, camera_class, roi_builder, self.result_writer, drawer_class=drawer_class,
            drawer_kwargs=drawer_kwargs,  user_data=self._user_data
        )
        self._add_submodule("recognizer", recognizer)

    def initialise_submodules(self):

        self.initialise_result_writer()

        if self.control:
            self.initialise_controller()

        if self.recognize:
            self.initialise_recognizer()

    @property
    def last_drawn_frame(self):
        if self.recognize:
            return self.recognizer.drawer.last_drawn_frame
        else:
            return None

    @property
    def last_t(self):
        if self.recognize and self.running:
            return self.recognizer.last_t

        return 0

    @property
    def clock(self):

        experiment_time = None
        adaptation_time = None

        if self.control:
            if self.controller is not None:

                adaptation_time = "%s / %s" % tuple([iso_format(hours_minutes_seconds(e)) for e in [
                    self.controller.adaptation_left,
                    self.controller.adaptation_offset
                ]])

            if self.controller.running:
                experiment_time = "%s / %s" % tuple([iso_format(hours_minutes_seconds(e)) for e in [
                    self.controller.elapsed_seconds,
                    self.controller.duration
                ]])


        clock = {
            "adaptation_time": adaptation_time,
            "experiment_time": experiment_time
        }

        return clock


    def run(self):
        """
        Update the submodules if required as long as the experiment can go on:
        * The elapsed seconds since this function was called is less than max_duration
        * Controller progress is les than 100%
        * The stop experiment method is not called

        If any of the three statements above stop are not True, the function terminates execution
        """
        super().run()

        # if self.recognize:
        #     self.recognizer.start_datetime = self.start_datetime
        #     self.recognizer.drawer.video_out = self.video_out

        if self.control:
            self.controller.adaptation_offset = self.adaptation_offset

        while not self.stopped and self.elapsed_seconds < self._max_duration:

            self._end_event.wait(1/self.sampling_rate)

            if self.control and self.controller is not None:
                self.controller.tick(self.last_t)
                if self.controller.progress > 100:
                    logger.info("Progress reached 100%")
                    self.stop_experiment()
                    break

        if self._user_data['run']:
            self.stop()

        return

    def stop(self):
        """
        Stop everything!
        """
        super().stop()
        self.stop_experiment(force=True)
        logger.info('Quitting control thread')


    def warm_up(self):
        """
        Convenience method to start the recognizer
        with the minimal controller hardware required to do an experiment
        This is listed in the _always_on_hardware list of the controller.
        """

        if self._warmed_up:
            warnings.warn("You already warmed up. Warming up again is disabled")
        else:
            self.controller.run_minimal()
            logger.debug("Preparing recognizer")
            self.recognizer.prepare()
            logger.debug("Starting recognizer")
            self.recognizer.start()
            self._warmed_up = True

    def start_experiment(self):
        """
        Convenience combination of commands targeted to start an experiment:

        * The start datetime of the experiment is set
        * The recognizer starts if it is not already
          (by) means of another method with the power of starting the recognizer
        * The controller starts executing its paradigm
        * The METADATA and ROI_MAP tables are generated
        """

        # set the self.start_datetime property to now with format
        # %YYYYY-MM-DD_HH-MM-SS
        self._set_start_datetime()

        if self._user_data['start_datetime'] is not None:
            self.result_writer.start_datetime = self._user_data['start_datetime']
        else:
            self.result_writer.start_datetime = self.start_datetime

        # allow the result_writer to write to disk the incoming data from now on
        self.result_writer.set_running()

        logger.warning("Saving results to %s", self.result_writer.result_dir)

        if self.recognize:
            self.recognizer.drawer.video_out = os.path.join(self._result_dir, self.machine_id, self.machine_name, self.start_datetime, self.run_id + '.avi')
            self.recognizer._time_running = True

            if not self.recognizer.ready:
                try:
                    # if the recognizer is not ready, it needs to prepare:
                    # * find the regions of interest (ROIS)
                    # * bind a tracker to each ROI

                    if self.controller is not None:
                        self.controller.run_minimal()

                    self.recognizer.prepare()
                except Exception as error:
                    logger.warning("Could not prepare recognizer. Please try again")
                    logger.warning(error)
                    logger.warning(traceback.print_exc())
                    return None

            if not self.recognizer.running:
                # start tracking animals
                self.recognizer.start()

        if self.control:
            # start the paradigm loaded in the controller
            self.controller.start()

        self.result_writer.initialise_metadata(self.metadata)
        if self.recognize and self.recognizer is not None:
            self.result_writer.initialise_roi_map(self.recognizer.rois)



    def stop_experiment(self, force=False):
        """
        Convenience combination of commands targeted to stop an experiment.

        * The recognizer stops tracking and recording frames
        * The controller stops doing experiments
        * The result writer clears its cache

        NOTE: This does not stop the execution of ControlThread.run().
        Only the stop() method can do that.
        """

        if not self._user_data["wrap"] or force:

            if self.recognize:
                if not self.recognizer.stopped:
                    self.recognizer.stop()

            if self.control:
                if not self.controller.stopped:
                    self.controller.stop()

            self.result_writer.clear()


    def reset_experiment(self):
        self.stop_experiment(force=True)
        Status.run(self)
        self.initialise_submodules()



r"""
Inform the user of the different paradigm.csv available,
load one of them into memory,
generate the parallel threads required to implement it in practice
"""

from collections import Counter
import logging
import os.path
from threading import Barrier

import numpy as np
import pandas as pd

from idoc.server.core.base import Settings

from idoc.configuration import IDOCConfiguration
from idoc.debug import IDOCException

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
config = IDOCConfiguration()

class Programmer(Settings):
    r"""
    Load user paradigms and prepare threads for concurrent control of an Arduino or Dummy board
    """

    table = None
    _paradigms_dir = None
    _required_columns = ["hardware", "start", "end", "on", "off", "mode", "value"]

    def __init__(
            self, *args, **kwargs
        ):

        super().__init__(*args, **kwargs)

        self._paradigms_dir = config.content["folders"]["paradigms"]["path"]
        self._barriers = {}
        self.table = []
        self._submodules = {}
        self._settings.update({
            'paradigm_path': None,
        })
        self._locked = False

    def list(self, dict_format=True):
        r"""
        Fetch a list of the paradigms available in the paradigm directory
        set in the config file
        This information is designed to be sent to the user.
        """
        files = sorted(os.listdir(self._paradigms_dir))

        if dict_format:
            return {i: f for i, f in enumerate(files) if f[::-1][:4][::-1] == ".csv"}
        else:
            return files

    @property
    def paradigm_name(self):
        return self._settings["paradigm_path"]

    @property
    def paradigm_path(self):
        if self._settings['paradigm_path'] is None:
            return None
        else:
            return os.path.join(self._paradigms_dir, self._settings["paradigm_path"])

    @paradigm_path.setter
    def paradigm_path(self, paradigm_path):
        r"""
        Load the user provided paradigm in csv format into Python
        and prepare the corresponding threads so the Controller
        can run them.
        """
        # load the data in paradigm_path into Python
        self._settings['paradigm_path'] = paradigm_path

        if not self._load_table(paradigm_path):
            self._settings['paradigm_path'] = config.content['controller']['paradigm_path']
            raise IDOCException('The provided paradigm could not be loaded. See more details below')


    @property
    def absolute_paradigm_path(self):
        if self._settings["paradigm_path"] is None:
            return None

        absolute_path = os.path.join(self._paradigms_dir, self._settings["paradigm_path"])
        if os.path.exists(absolute_path) and os.path.isfile(absolute_path):
            pass
        else:
            logger.warning('The passed paradigm does not exist. Check it is correct!!:')
            logger.warning(absolute_path)

        return absolute_path

    def _validate_paradigm_df(self, df):
        """
        Check the data frame provided in the paradigm_path .csv file
        fulfills the IDOC requirements

        * It should contain the columns in self._required_columns
        """
        # TODO Add more checks

        missing_columns = []
        for column in self._required_columns:
            if column not in df.columns:
                missing_columns.append(column)

        if len(missing_columns) != 0:
            logger.warning("%s are missing in the paradigm. Please add them", ' '.join(missing_columns))
            return False

        return True

    def _validate_paradigm_path(self, paradigm_path):
        """
        Check the paradigm path is valid. It should:

        * Not be None
        * Be within the list of available paradigms
        """
        if paradigm_path is None:

            self.table = None
            return False

        elif paradigm_path not in self.list(dict_format=False):
            self.table = None
            logger.warning("The passed paradigm does not exist!")
            return False

        else:
            return True

    def _load_table(self, paradigm_path):
        """
        Load a csv into python in list format
        Each row becomes a dictionary inside the list
        and the columns and their values are encoded
        as key-value pairs of the dictionary.
        """

        # make sure the old table is reset, if populated at all
        self.table.clear()

        # series of checks to make sure the provided paradigm is valid
        if not self._validate_paradigm_path(paradigm_path):
            logger.warning("Error reading your paradigm into IDOC. Paradigm path is invalid")
            return False

        table_df = pd.read_csv(self.absolute_paradigm_path, comment='#')
        if not self._validate_paradigm_df(table_df):
            logger.warning("Error reading your paradigm into IDOC. Paradigm is invalid")
            return False


        # convert on and off from ms to seconds
        # we assume the user enters this in ms because its more convenient for him/her
        table_df["start"] *= 60
        table_df["end"] *= 60
        table_df["on"] /= 1000
        table_df["off"] /= 1000

        i = 0
        # coerce to a dictionary
        for row_tuple in table_df.iterrows():
            row = row_tuple[1].to_dict()
            self.table.append(row)
            i += 1

        if self.table == []:
            logger.warning("Error reading your paradigm into IDOC. Table is empty")
            return False

        # print(self.table)

        return True

    @staticmethod
    def parse(row, i):
        """
        Process the content of the row inside prepare() and return a
        kwargs dictionary that can be unpacked upon BaseControllerThead
        initializations.
        """

        kwargs = {
            "i": str(i).zfill(3),
            "hardware": row['hardware'],
            "start": row["start"],
            "end": row["end"],
            "value": float(row['value']),
            "mode": row['mode']
        }

        # Decide here which thread to use
        # Either a thread that leaves the pin on constitutively
        if np.isnan(row['on']) and np.isnan(row['off']):
            thread = "DEFAULT"

        # or a thread that implements a duty cycle with some frequency
        # and phase given by on and off
        elif not np.isnan(row['on']) and not np.isnan(row['off']):
            thread = "WAVE"
            kwargs["seconds_on"] = row["on"]
            kwargs["seconds_off"] = row["off"]

        else:
            raise IDOCException('Row %d: one of on or off is defined, but not both. Please fix this' % i)

        return thread, kwargs

    @staticmethod
    def make_barriers(table):
        r"""
        Generate barriers for perfect on-off alternation between threads
        using the start and stop times of the threads in the paradigm
        Make them so there is one and only one for every timepoint
        where at least one thread either stops or starts
        The barrier will expect as many threads
        as the sum of starting and stopping threads
        The threads are given a start and a stop barrier
        which is given so its shared across threads
        with equal endpoint (sharing the end barrier)
        or starting right after (the end barrier is the same as the start barrier of those).
        """

        barriers = {}

        switch_times = {
            "start": Counter([e["start"] for e in table]),
            "end": Counter([e["end"] for e in table]),
            "total": Counter([v[e] for e in ["start", "end"] for v in table])
        }

        barriers = {k: Barrier(v) for k, v in switch_times["total"].items()}

        for switch_time in barriers:
            logger.debug(
                "Barrier at time %d with id %s will wait for %d parties",
                switch_time,
                id(barriers[switch_time]),
                barriers[switch_time].parties
            )

        return barriers

    @staticmethod
    def add_barriers(paradigm, barriers):
        r"""
        Add the barriers to the threads in the right position.
        """

        # provide the threads with barriers for improved synchronization
        # at the expense of slightly increased waiting times
        for thread in paradigm:

            start_barrier = barriers[thread.start_seconds]
            thread.add_barrier(start_barrier, "start")

            end_barrier = barriers[thread.end_seconds]
            thread.add_barrier(end_barrier, "end")

        return paradigm

    def lock(self):
        self._locked = True



__author__ = 'quentin'

import logging
import itertools
import pickle
import os
import yaml

import numpy as np
import cv2


from idoc.server.roi_builders.roi_builders import BaseROIBuilder
from idoc.server.core.roi import ROI
from idoc.debug import IDOCException, EthoscopeException
from idoc.server.roi_builders.helpers import find_quadrant
from idoc.server.roi_builders.helpers import place_dots
from idoc.configuration import IDOCConfiguration

try:
    CV_VERSION = int(cv2.__version__.split(".")[0]) # pylint: disable=no-member
except:
    CV_VERSION = 2

try:
    from cv2.cv import CV_CHAIN_APPROX_SIMPLE as CHAIN_APPROX_SIMPLE # type: ignore
    from cv2.cv import CV_AA as LINE_AA  # type: ignore
except ImportError:
    from cv2 import CHAIN_APPROX_SIMPLE
    from cv2 import LINE_AA

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)

config = IDOCConfiguration()

class TargetGridROIBuilder(BaseROIBuilder):

    _adaptive_med_rad = 0.10
    _expected_min_target_dist = 200 # the minimal distance between two targets, in 'target diameter'
    _n_rows = 10
    _n_cols = 2
    _top_margin = 0
    _bottom_margin = None
    _left_margin = 0
    _right_margin = None
    _horizontal_fill = 1
    _vertical_fill = None
    _inside_pad = 0.0
    _outside_pad = 0.0
    _reverse = False
    # MAGIC NUMBERS
    _max_non_zero = 0.7
    _min_area_perimeter_ratio = 0.8
    _min_target_area = 400
    _max_target_area = 800
    _min_area_ratio = 1
    _max_area_ratio = 1.5
    # END

    _rois_pickle_file = "rois.pickle"

    _description = {"overview": "A flexible ROI builder that allows users to select parameters for the ROI layout."
                               "Lengths are relative to the distance between the two bottom targets (width)",
                    "arguments": [
                        {"type": "number", "min": 1, "max": 16, "step":1, "name": "n_cols", "description": "The number of columns", "default":1},
                        {"type": "number", "min": 1, "max": 16, "step":1, "name": "n_rows", "description": "The number of rows", "default":1},
                        {"type": "number", "min": 0.0, "max": 1.0, "step":.001, "name": "top_margin", "description": "The vertical distance between the middle of the top ROIs and the middle of the top target.", "default":0.0},
                        {"type": "number", "min": 0.0, "max": 1.0, "step":.001, "name": "bottom_margin", "description": "Same as top_margin, but for the bottom.", "default":0.0},
                        {"type": "number", "min": 0.0, "max": 1.0, "step":.001, "name": "right_margin", "description": "Same as top_margin, but for the right.", "default":0.0},
                        {"type": "number", "min": 0.0, "max": 1.0, "step":.001, "name": "left_margin", "description": "Same as top_margin, but for the left.", "default":0.0},
                        {"type": "number", "min": 0.0, "max": 1.0, "step":.001, "name": "horizontal_fill", "description": "The proportion of the grid space used by the roi, horizontally.", "default":0.90},
                        {"type": "number", "min": 0.0, "max": 1.0, "step":.001, "name": "vertical_fill", "description": "Same as horizontal_margin, but vertically.", "default":0.90}
                    ]}

    def __init__(self, *args, n_rows=1, n_cols=1, top_margin=0, bottom_margin=0,
                 left_margin=0, right_margin=0, horizontal_fill=.9, vertical_fill=.9, inside_pad=.0, outside_pad=.0, **kwargs):
        """
        This roi builder uses three black circles drawn on the arena (targets) to align a grid layout:

        IMAGE HERE

        :param n_rows: The number of rows in the grid.
        :type n_rows: int
        :param n_cols: The number of columns.
        :type n_cols: int
        :param top_margin: The vertical distance between the middle of the top ROIs and the middle of the top target
        :type top_margin: float
        :param bottom_margin: same as top_margin, but for the bottom.
        :type bottom_margin: float
        :param left_margin: same as top_margin, but for the left side.
        :type left_margin: float
        :param right_margin: same as top_margin, but for the right side.
        :type right_margin: float
        :param horizontal_fill: The proportion of the grid space user by the roi, horizontally (between 0 and 1).
        :type horizontal_fill: float
        :param vertical_fill: same as vertical_fill, but horizontally.
        :type vertical_fill: float
        """

        self._n_rows = n_rows
        self._n_cols = n_cols
        self._top_margin =  top_margin
        self._bottom_margin = bottom_margin
        self._left_margin = left_margin
        self._right_margin = right_margin
        self._horizontal_fill = horizontal_fill
        self._vertical_fill = vertical_fill
        self._inside_pad = inside_pad
        self._outside_pad = outside_pad
        self._sorted_src_pts = [None, ] * 3
        self._img = None

        if 'rois_pickle_file' in kwargs.keys():
            self._rois_pickle_file = kwargs.pop('rois_pickle_file')

        super().__init__(*args, **kwargs)

    def _sort_src_pts(self, src_points):
        """
        Sort the three dots / targets on the arena by their position
        Maximum distance is between A and C
        The point not having the most distance with other points is thus B
        C is the one with the highest distance to B
        A is the other one (with the lowest distance)

        """
        a, b, c = src_points

        pairs = [(a, b), (b, c), (a, c)]

        dists = [self._points_distance(*p) for p in pairs]

        for d in dists:
            logger.info(d)
            if d < self._expected_min_target_dist:
                raise EthoscopeException(
                                    """
                                    The detected targets are too close.
                                    If you think this is correct, please decrease the value of _expected_min_target_dist
                                    to something that fits your setup. It is set to %i
                                    """ % self._expected_min_target_dist
                                    )
        # that is the AC pair
        hypo_vertices = pairs[np.argmax(dists)]

        # this is B : the only point not in (a,c)
        for sp in src_points:
            if not sp in hypo_vertices:
                break
        sorted_b = sp # pylint: disable=undefined-loop-variable

        dist = 0
        for sp in src_points:
            if sorted_b is sp:
                continue
            # b-c is the largest distance, so we can infer what point is c
            if self._points_distance(sp, sorted_b) > dist:
                dist = self._points_distance(sp, sorted_b)
                sorted_c = sp

        # the remaining point is a
        sorted_a = [sp for sp in src_points if not sp is sorted_b and not sp is sorted_c][0]
        sorted_src_pts = np.array([sorted_a, sorted_b, sorted_c], dtype=np.float32)
        if self._reverse:
            sorted_src_pts = sorted_src_pts[::-1]

        self._sorted_src_pts = sorted_src_pts
        logger.debug(sorted_src_pts)

        return sorted_src_pts

    def _find_blobs(self, image, scoring_fun=None):

        if scoring_fun is None:
            scoring_fun = self._score_targets

        grey = image
        radius = int(self._adaptive_med_rad * image.shape[1])
        if radius % 2 == 0:
            radius += 1

        # cv2.imhow('grey', grey)
        # cv2.waitKey(0)

        med = np.median(grey)
        if med > 20:
            scale = 255 / (med)
            cv2.multiply(grey, scale, dst=grey)
            binary = np.copy(grey)
        else:
            binary = np.copy(grey)

        # cv2.imshow('bin', bin)
        # cv2.waitKey(0)
        score_map = np.zeros_like(binary)
        for threshold in range(0, 255, 5):

            cv2.threshold(grey, threshold, 255, cv2.THRESH_BINARY, binary)
            if np.count_nonzero(binary) > self._max_non_zero * image.shape[0] * image.shape[1]:
                continue
            if CV_VERSION == 3:
                _, contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)
            else:
                contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)

            binary.fill(0)
            for contour in contours:
                score = scoring_fun(contour)
                if score > 0:
                    cv2.drawContours(binary, [contour], 0, score, -1)

            cv2.add(binary, score_map, score_map)

        return score_map

    def _find_blobs_new(self, image, scoring_fun=None):

        grey = image
        if scoring_fun is None:
            scoring_fun = self._score_circles

        # radius of dots is 0.1 * width of image
        # units?
        rad = int(self._adaptive_med_rad * grey.shape[1])
        # make sure it is an odd
        if rad % 2 == 0:
            rad += 1

        # make median intensity 255
        med = np.median(grey)
        scale = 255/(med)
        cv2.multiply(grey, scale, dst=grey)
        binary = np.copy(grey)
        score_map = np.zeros_like(binary)
        # grey_orig = grey.copy()

        for threshold in range(0, 255, 5):
            cv2.threshold(grey, threshold, 255, cv2.THRESH_BINARY_INV, binary)

            non_zero_pixels = np.count_nonzero(binary)

            # if this threshold operation leaves
            # more than 70% white pixels, continue
            # until loop is over
            if non_zero_pixels > self._max_non_zero * grey.shape[0] * grey.shape[1]:
                continue

            if CV_VERSION == 3:
                _, contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)
            else:
                contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)

            foreground_model = cv2.drawContours(grey.copy(), contours, -1, 255, -1)
            _, foreground_model = cv2.threshold(foreground_model, 254, 255, cv2.THRESH_BINARY)
            foreground_model = cv2.morphologyEx(foreground_model, cv2.MORPH_CLOSE, kernel=np.ones((3, 3)), iterations=1)

            if CV_VERSION == 3:
                _, contours, _ = cv2.findContours(foreground_model, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)
            else:
                contours, _ = cv2.findContours(foreground_model, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)

            binary.fill(0)

            for contour in contours:
                score = scoring_fun(contour)
                if score > 0:
                    cv2.drawContours(binary, [contour], 0, score, -1)

            score_map = cv2.add(binary, score_map, score_map)


        return score_map

    @staticmethod
    def _make_grid(
            n_col, n_row,
            top_margin=0.0, bottom_margin=0.0,
            left_margin=0.0, right_margin=0.0,
            horizontal_fill=1.0, vertical_fill=1.0,
            inside_pad=0.0, outside_pad=0.0
        ):

        y_positions = (np.arange(n_row) * 2.0 + 1) * (1 - top_margin - bottom_margin) / (2 * n_row) + top_margin
        x_positions = (np.arange(n_col) * 2.0 + 1) * (1 - left_margin - right_margin) / (2 * n_col) + left_margin
        all_centres = [np.array([x, y]) for x, y in itertools.product(x_positions, y_positions)]

        padded_centres = []
        for center in all_centres:
            left, top = find_quadrant((1, 1), center)
            if left:
                center[0] -= inside_pad
                center[0] += outside_pad
            else:
                center[0] += inside_pad
                center[0] -= outside_pad

            padded_centres.append(center)
        all_centres = padded_centres

        sign_mat = np.array([
            [-1, -1],
            [+1, -1],
            [+1, +1],
            [-1, +1]

        ])
        xy_size_vec = np.array([horizontal_fill/float(n_col), vertical_fill/float(n_row)]) / 2.0
        rectangles = [sign_mat *xy_size_vec + c for c in all_centres]
        return rectangles


    @staticmethod
    def _points_distance(pt1, pt2):
        x1 , y1  = pt1
        x2 , y2  = pt2
        return np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

    def _score_targets(self, contour):

        area = cv2.contourArea(contour)
        perimeter = cv2.arcLength(contour, True)

        if perimeter == 0:
            return 0
        circle = 4 * np.pi * area / perimeter ** 2

        if circle < self._min_area_perimeter_ratio:
            return 0

        return 1

    def _score_circles(self, contour):
        """
        Score the target quality by checking
        * its area is within limits
        * the ratio of the area of the minimal enclosing circle and its area is within limits
        """

        area = cv2.contourArea(contour)

        if not self._min_target_area > area > self._max_target_area:
            return 0

        ((x, y), radius) = cv2.minEnclosingCircle(contour)
        # if area was zero the functionw would have already returned
        ratio = np.pi * radius ** 2 / area

        if self._min_area_ratio < ratio < self._max_area_ratio:
            return 1

        return 0

    def _find_target_coordinates(self, image, blob_function):
        """
        :param image: A single channel image (2D)
        :type image: np.array
        """

        if len(image.shape) != 2:
            raise IDOCException("find_target_coordinates: image is not 2D")

        score_map = blob_function(image)

        # cv2.imshow('blob_score_map', score_map)
        # cv2.waitKey(0)

        binary = np.zeros_like(score_map)

        # as soon as we have three objects, we stop
        contours = []
        for threshold in range(0, 255, 1):

            cv2.threshold(score_map, threshold, 255, cv2.THRESH_BINARY, binary)

            # cv2.imshow('threshold', binary)
            # cv2.waitKey(0)

            if CV_VERSION == 3:
                _, contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)
            else:
                contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)

            if len(contours) < 3:
                raise EthoscopeException("There should be three targets. Only %i objects have been found" % (len(contours)), image)
            if len(contours) == 3:
                break

        target_diams = [cv2.boundingRect(contour)[2] for contour in contours]

        mean_diam = np.mean(target_diams)
        mean_sd = np.std(target_diams)

        if mean_sd / mean_diam > 0.10:
            raise EthoscopeException(
                """
                Too much variation in the diameter of the targets.
                Something must be wrong since all target should have the same size
                """,
                image
            )

        src_points = []
        for contour in contours:
            moms = cv2.moments(contour)
            x, y = moms["m10"] / moms["m00"], moms["m01"] / moms["m00"]
            src_points.append((x, y))

        sorted_src_pts = self._sort_src_pts(src_points)

        image_dots = place_dots(image.copy(), sorted_src_pts, color=0)
        cv2.imwrite(os.path.join("image_dots.png"), image_dots)

        return sorted_src_pts

    def _rois_from_img(self, image):
        """
        :param image: A 3 channel image. It can be gray, but 3 channels are needed
        :type image: np.array
        """
        self._img = image

        # Part 0: if a pickle file storing the position of the rois is provided
        # just load it and return the rois contained in it, terminating the routine
        if os.path.exists(self._rois_pickle_file):
            with open(self._rois_pickle_file, "rb") as filehandle:
                rois = pickle.load(filehandle)
            return rois

        else:
            # Part 1.A if there is no pickle file but a target_coord_file is available
            # load the sorted_src_pts from it and continue to part 2
            sorted_src_pts = None
            target_coord_file = config.content["roi_builder"]["target_coord_file"]


            if target_coord_file != "None":
                if os.path.exists(target_coord_file):
                    try:
                        with open(target_coord_file, "r") as filehandle:
                            data = yaml.load(filehandle, yaml.SafeLoader)
                    
                        logger.info("Reading manually entered coordinates of targets")
                        # each dot is made by two numbers separated by comma
                        src_points = [tuple(v) for v in data.values()]
                        sorted_src_pts = self._sort_src_pts(src_points)
                    except Exception as error:
                        logger.warning(f"Could not load yaml file {target_coord_file}. Please make sure it is valid!")
                        logger.warning(error)
                else:
                    logger.warning(f"Could not load yaml file {target_coord_file}. Please make sure it exists")


            # Part 1.B no pickle file nor target_coord_file is present. Try finding where the targets are
            # calling find_target_coordinates
            if sorted_src_pts is None:
                if len(self._img.shape) == 3 and self._img.shape[2] == 3:
                    img = cv2.cvtColor(self._img, cv2.COLOR_BGR2GRAY)
                else:
                    img = self._img
                try:
                    sorted_src_pts = self._find_target_coordinates(img, self._find_blobs)
                except EthoscopeException as e:
                    # raise e
                    logger.warning("Fall back to find_blobs_new")
                    sorted_src_pts = self._find_target_coordinates(img, self._find_blobs_new)
                except Exception as e:
                    raise e

            # Part 2 Fit a grid to the frame using the three detected targets
            # Use the grid to position the ROIs in the right place
            # and return them as a list of objects of class ROI.

            dst_points = np.array([
                (0, -1),
                (0, 0),
                (-1, 0)
            ], dtype=np.float32)

            wrap_mat = cv2.getAffineTransform(dst_points, sorted_src_pts)

            rectangles = self._make_grid(
                self._n_cols, self._n_rows,
                self._top_margin, self._bottom_margin,
                self._left_margin, self._right_margin,
                self._horizontal_fill, self._vertical_fill,
                self._inside_pad, self._outside_pad
            )

            shift = np.dot(wrap_mat, [1, 1, 0]) - sorted_src_pts[1] # point 1 is the ref, at 0,0
            rois = []

            for i, r in enumerate(rectangles):

                r = np.append(r, np.zeros((4, 1)), axis=1)
                mapped_rectangle = np.dot(wrap_mat, r.T).T
                mapped_rectangle -= shift
                ct = mapped_rectangle.astype(np.int32)
                cv2.drawContours(self._img, [ct.reshape((1, 4, 2))], -1, (255, 0, 0), 1, LINE_AA)
                rois.append(ROI(ct, idx=i+1))

        return rois


class FSLSleepMonitorWithTargetROIBuilder(TargetGridROIBuilder):

    _description = {"overview": "The default sleep monitor arena with ten rows of two tubes.",
                    "arguments": []}

    def __init__(self, *args, **kwargs):
        r"""
        Class to build ROIs for a two-columns, ten-rows for the sleep monitor
        (`see here <https://github.com/gilestrolab/ethoscope_hardware/tree/master/arenas/arena_10x2_shortTubes>`_).
        """
        #`sleep monitor tube holder arena <todo>`_

        super().__init__(
            *args,
            n_rows=10,
            n_cols=2,
            top_margin=6.99 / 111.00,
            bottom_margin=6.99 / 111.00,
            left_margin=-.033,
            right_margin=-.033,
            horizontal_fill=.8,
            vertical_fill=.5,
            inside_pad=0.07,
            outside_pad=0.0,
            **kwargs
        )



class IDOCROIBuilder(TargetGridROIBuilder):

    _reverse = True

    def __init__(self, *args, **kwargs):

        super().__init__(
            *args,
            n_rows=10, n_cols=2,
            top_margin=2.3 / 111.00, bottom_margin=1.8 / 111.00,
            left_margin=0.033, right_margin=0.033,
            horizontal_fill=.5, vertical_fill=.40,
            inside_pad=.05, outside_pad=.0,
            **kwargs
        )



import json
import logging
import urllib.request

from idoc.decorators import retry
from idoc.debug import ScanException

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class HTTPMixin:

    @retry(ScanException, tries=3, delay=1, backoff=1)
    def _get_json(self, url, timeout=5, post_data=None):

        try:
            req = urllib.request.Request(url, data=post_data, headers={'Content-Type': 'application/json'})
            file_handle = urllib.request.urlopen(req, timeout=timeout)
            message = file_handle.read()
            if not message:
                # logging.error("URL error whist scanning url: %s. No message back." % self._id_url)
                raise ScanException("No message back")
            try:
                resp = json.loads(message)
                return resp
            except ValueError:
                # logging.error("Could not parse response from %s as JSON object" % self._id_url)
                raise ScanException("Could not parse Json object")

        except urllib.error.HTTPError as error:
            logger.warning('Cannot open URL %s', url)
            raise ScanException("Error" + str(error.code))
            #return e

        except urllib.error.URLError as error:
            logger.warning('Cannot open URL %s', url)
            raise ScanException("Error" + str(error.reason))
            #return e

        except Exception as error:
            raise ScanException("Unexpected error" + str(error))



import cv2
import numpy as np
from idoc.debug import IDOCException

__author__ = 'quentin'


class ROI(object):

    def __init__(self, polygon, idx, value=None, orientation=None, regions=None, side=None):
        """
        Class to define a region of interest(ROI).
        Internally, ROIs are single polygons.
        At the moment, they cannot have any holes.
        The polygon defining the ROI is used to draw a mask to exclude off-target pixels (so cross-ROI info).

        :param polygon: An array of points
        :type polygon: :class:`~numpy.ndarray`
        :param idx: the index of this ROI
        :type idx: int
        :param value: an optional value to be save for this ROI (e.g. to define left and right side)
        :param orientation: Optional orientation Not implemented yet
        :param regions: Optional sub-regions within the ROI. Not implemented yet

        """

        # TODO if we do not need polygon, we can drop it
        self.side = side
        self._polygon = np.array(polygon)
        if len(self._polygon.shape) == 2:
            self._polygon = self._polygon.reshape((self._polygon.shape[0],1,self._polygon.shape[1]))


        x,y,w,h = cv2.boundingRect(self._polygon)

        self._mask = np.zeros((h,w), np.uint8)
        cv2.drawContours(self._mask, [self._polygon], 0, 255,-1,offset=(-x,-y))

        self._rectangle = x,y,w,h
        # todo NOW! sort rois by value. if no values, left to right/ top to bottom!
        self._idx = idx

        if value is None:
            self._value = self._idx
        else:
            self._value = value

    @property
    def idx(self):
        """
        :return: The index of this ROI
        :rtype: int
        """
        return self._idx

    def bounding_rect(self):
        raise NotImplementedError


    def mask(self):
        """
        :return: The mask as a single chanel, `uint8` image.
        :rtype: :class:`~numpy.ndarray`
        """
        return self._mask

    @property
    def offset(self):
        """
        :return: the x,y offset of the ROI compared to the frame it was build on.
        :rtype: (int,int)
        """
        x,y,w,h = self._rectangle
        return x,y

    @property
    def polygon(self):
        """
        :return: the internal polygon defining the ROI.
        :rtype: :class:`~numpy.ndarray`
        """
        return self._polygon


    @property
    def longest_axis(self):
        """
        :return: the value of the longest axis (w or h)
        :rtype: float
        """
        x,y,w,h = self._rectangle
        return float(max(w, h))

    @property
    def rectangle(self):
        """
        :return: The upright bounding rectangle to the ROI formatted (x,y,w,h). Where x and y are to coordinates of the top left corner
        :rtype: (int,int,int,int)
        """
        return self._rectangle

    def get_feature_dict(self):
        """
        :return: A dictionary of freatures for this roi. It containes the folowing fields:

        * "x"
        * "y"
        * "w"
        * "h"
        * "value"
        * "idx"

        :rtype: dict
        """
        x,y,w,h = self._rectangle
        return {"x":x,
                "y":y,
                "w":w,
                "h":h,
                "value":self._value,
                "idx":self.idx
        }


    def set_value(self, new_val):
        """
        :param new_val: assign a nex value to a ROI
        """
        self._value = new_val

    @property
    def value(self):
        """
        :return: the value of a ROI
        """
        return self._value

    def apply(self,img):
        """
        Cut an image where the ROI is defined.

        :param img: An image. Typically either one or three channels `uint8`.
        :type img: :class:`~numpy.ndarray`
        :return: a tuple containing the resulting cropped image and the associated mask (both have the same dimension).
        :rtype: (:class:`~numpy.ndarray`, :class:`~numpy.ndarray`)
        """
        x,y,w,h = self._rectangle

        try:
            out = img[y : y + h, x : x +w]
        except:
            raise IDOCException("Error whilst slicing region of interest %s" % str(self.get_feature_dict()), img)

        if out.shape[0:2] != self._mask.shape:
            raise IDOCException("Error whilst slicing region of interest. Possibly, the region out of the image: %s" % str(self.get_feature_dict()), img )

        return out, self._mask


# https://stackoverflow.com/a/4761058
import copy
import logging
from pathlib import Path
import sys
import unittest

from . import test_generic
from LeMDT import PROJECT_DIR
from LeMDT.LeMDT.interface import Interface
from LeMDT.LeMDT.interface import Tracker
from pypylon import pylon, genicam


class TestProgramLoading(unittest.TestCase, Interface):

    def __init__(self, *args, **kwargs):

        interface = Interface(
             arduino = True, port = '/dev/ttyACM0',
             track = True, camera = 'pylon',
             duration = 1, gui = 'tkinter'
        )
        
        interface.init_components()
        self.interface = interface
        unittest.TestCase.__init__(self, *args, **kwargs)


    def test_loading_program_updates_threads(self):
        """
        Make sure that selecting a new file
        regenerates the threads i.e. we dont stay
        with the defaults
        """
        program_path = Path(PROJECT_DIR, 'programs', 'test_28_march.csv').as_posix()
        self.interface.device.program_path = program_path

        before_threads = {**self.interface.device.threads}
        after_threads = self.interface.gui.load_program(program_path)       
        self.assertNotEqual(before_threads, after_threads)



import unittest
from idoc.server.controllers.controllers import Controller



class TestController(unittest.TestCase):

    _mapping_path = "/1TB/Cloud/Lab/Gitlab/idoc/py_src/mappings/mega.csv"

    def setUp(self):

        self.controller = Controller(mapping_path=self._mapping_path, program_path=None, board="ArduinoDummy", port="")

    def test_load_mapping(self):

        self.controller.load_mapping(self._mapping_path)
        self.assertTrue(len(self.controller._mapping) > 1)



import threading

def or_set(self):
    self._set()
    self.changed()

def or_clear(self):
    self._clear()
    self.changed()

def orify(e, changed_callback):
    e._set = e.set
    e._clear = e.clear
    e.changed = changed_callback
    e.set = lambda: or_set(e)
    e.clear = lambda: or_clear(e)

def OrEvent(*events):
    or_event = threading.Event()
    def changed():
        bools = [e.is_set() for e in events]
        if any(bools):
            or_event.set()
        else:
            or_event.clear()
    for e in events:
        orify(e, changed)
    changed()
    return or_event


__author__ = 'quentin'


import logging
import traceback
import os.path

import cv2
import numpy as np

from idoc.server.core.base import DescribedObject
from idoc.server.core.roi import ROI
from idoc.debug import IDOCException, EthoscopeException

logger = logging.getLogger(__name__)

class BaseROIBuilder(DescribedObject):

    def __init__(self):
        """
        Template to design ROIBuilders. Subclasses must implement a ``_rois_from_img`` method.
        """

    @staticmethod
    def fetch_frames(input):
        # If input is an image, make a copy
        # Otherwise it is assumed to be a camera object
        # that returns frames upon iterating it.
        # Capture 5 frames and take the median intensity for each pixel
        # i.e. get an image that represents the median of 5 frames
        accum = []
        logger.debug("Fetching frames")

        if isinstance(input, np.ndarray):
            accum = np.copy(input)
        else:
            for idx, (_, frame) in enumerate(input):
                logger.debug(frame.shape)
                accum.append(frame)
                if idx == 5:
                    break

            accum = np.median(np.array(accum), 0).astype(np.uint8)

        output_path = os.path.join(os.environ["HOME"], "input.png")
        logger.info("Saving accum to %s", output_path)
        cv2.imwrite(output_path, accum)

        return accum

    def build(self, input):
        """
        Uses an input (image or camera) to build ROIs.
        When a camera is used, several frames are acquired and averaged to build a reference image.

        :param input: Either a camera object, or an image.
        :type input: :class:`~ethoscope.hardware.input.camera.BaseCamera` or :class:`~numpy.ndarray`
        :return: list(:class:`~ethoscope.core.roi.ROI`)
        """

        accum = self.fetch_frames(input)

        try:
            rois = self._rois_from_img(accum)
            img = accum

            # try:
            #     for pt in self._sorted_src_pts:
            #         roi_build_with_dots = img.copy()
            #         roi_build_with_dots = cv2.circle(roi_build_with_dots, tuple(pt), 5, (0,255,0), -1)

            #     cv2.imwrite(os.path.join(os.environ["HOME"], "roi_build_with_dots.png"), roi_build_with_dots)
            # except AttributeError as e:
            #     logging.warning(e)

        except Exception as e:
            if not isinstance(input, np.ndarray):
                del input
            logging.error(traceback.format_exc())
            raise e

        rois_w_no_value = [r for r in rois if r.value is None]

        if len(rois_w_no_value) > 0:
            rois = self._spatial_sorting(rois)
        else:
            rois = self._value_sorting(rois)
        return rois

    def _rois_from_img(self, input=None):
        raise NotImplementedError

    def _spatial_sorting(self, rois):
        '''
        returns a sorted list of ROIs objects it in ascending order based on the first value in the rectangle property
        '''
        return sorted(rois, key=lambda x: x.rectangle[0], reverse=False)

    def _value_sorting(self, rois):
        '''
        returns a sorted list of ROIs objects it in ascending order based on the .value property
        '''
        return sorted(rois, key=lambda x: x.value, reverse=False)


class DefaultROIBuilder(BaseROIBuilder):


    """
    The default ROI builder. It simply defines the entire image as a unique ROI.
    """
    _description = {"overview": "The default ROI builder. It simply defines the entire image as a unique ROI.", "arguments": []}


    def _rois_from_img(self,img):
        h, w = img.shape[0],img.shape[1]
        return[
            ROI(np.array([
                (   0,        0       ),
                (   0,        h -1    ),
                (   w - 1,    h - 1   ),
                (   w - 1,    0       )])
            , idx=1)]




import argparse
import logging
import signal
import time
from pyfirmata import ArduinoMega as Arduino

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


ap = argparse.ArgumentParser()
ap.add_argument("-p", "--port", type = str)
ap.add_argument("-v", "--value", type = float, help = "0-1")
ap.add_argument('--pins', nargs='+', help='List of pins', required=True)
ap.add_argument('--hertz', type=float)

args = ap.parse_args()
args = vars(args)

def main():

    board = Arduino(args["port"])
    stop = False

    def quit(signo=None, _frame=None):
        nonlocal stop

        logger.info("Received signal %s", signo)
        stop = True
        [board.digital[int(p)].write(0) for p in args["pins"]]

    signals = ('TERM', 'HUP', 'INT')
    for sig in signals:
        signal.signal(getattr(signal, 'SIG' + sig), quit)


    VALUE = args['value']
    VALUE = 1 if VALUE == 1.0 else VALUE

    PINS = {}
    for p in args['pins']:
        if isinstance(VALUE, float):
            PINS[p] = board.get_pin('d:%d:p' % int(p))
        else:
            PINS[p] = board.get_pin('d:%d:o' % int(p))


    if args['hertz'] is None:
        [PINS[p].write(args["value"]) for p in PINS]
        while not stop:
            print('Waiting for you to stop me... (Control-C)')
            time.sleep(1)

        if stop:
            print('Turning off normally...')

    else:
        while not stop:
            [p.write(VALUE) for p in PINS.values()]
            time.sleep(1/(args['hertz']*2))
            [p.write(0) for p in PINS.values()]
            time.sleep(1/(args['hertz']*2))
            print('Waiting for you to stop me... (Control-C)')

        if stop:
            print('Turning off normally...')




__author__ = 'antonio'

import logging

import cv2
import numpy as np

from idoc.server.core.base import Settings
from idoc.server.roi_builders.roi_builders import BaseROIBuilder
from idoc.server.core.roi import ROI
from idoc.debug import EthoscopeException

try:
    CV_VERSION = int(cv2.__version__.split(".")[0])
except:
    CV_VERSION = 2

try:
    from cv2.cv import CV_CHAIN_APPROX_SIMPLE as CHAIN_APPROX_SIMPLE # type: ignore
    from cv2.cv import CV_AA as LINE_AA  # type: ignore
except ImportError:
    from cv2 import CHAIN_APPROX_SIMPLE
    from cv2 import LINE_AA

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class HighContrastROIBuilder(BaseROIBuilder, Settings):

    _adaptive_med_rad = 0.05
    _expected__min_target_dist = 10 # the minimal distance between two targets, in 'target diameter'
    _description = {
        "overview": """
        A flexible ROI builder that allows users to select parameters for the ROI layout.
        Lengths are relative to the distance between the two bottom targets (width)
        """
    }

    def __init__(self, *args, nrows=10, ncols=2, **kwargs):
        """
        This roi builder takes advantage of the hight contrast
        between the ROIs and the background when illuminated with IR light.
        It is particularly powerful when the ROI is the only area with light.

        :param nrows: The number of rows in the grid.
        :type nrows: int
        :param ncols: The number of columns.
        :type ncols: int
        """

        self._nrows = nrows
        self._ncols = ncols
        self._min_area = 1500
        self._max_area = 2500

        self._min_thresh = 0
        self._max_thresh = 255
        self._inter_thresh = 4

        super().__init__(*args, **kwargs)


    def _find_contours(self, binary_image):
        r"""
        Receive a binary image and return a list of contours
        """

        while np.count_nonzero(binary_image) > 0:

            # bin = cv2.morphologyEx(bin, cv2.MORPH_TOPHAT, (9, 9))
            binary_image = cv2.erode(binary_image, np.ones((1, 10)))

            # cv2.imshow('binary', binary_image)
            # cv2.waitKey(0)

            if CV_VERSION == 3:
                _, contours, _ = cv2.findContours(binary_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            else:
                contours, _ = cv2.findContours(binary_image, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)

            if len(contours) == self._nrows * self._ncols:
                return contours

        # cv2.imwrite(os.path.join(os.environ["HOME"], 'failed_roi_builder_binary.png'), bin_orig)
        raise EthoscopeException("")

    @staticmethod
    def _find_center(contour):
        r"""
        Return the center of the bounding box of the contour.
        """
        x, y, w, h = cv2.boundingRect(contour)
        width = np.max([w, h])
        height = np.min([w, h])
        center = (x + width // 2, y + height // 2)
        return center

    @staticmethod
    def _find_quadrant(shape, center):
        r"""
        Return a str of length 2 encoding wich quadrant
        the center is in:
        tl, tr, br, bl
        """
        quadrant = ['b', 'r']

        left = center[0] < (shape[1] / 2)
        top = center[1] > (shape[0] / 2)
        if top:
            quadrant[0] = 't'
        if left:
            quadrant[1] = 'l'

        return ''.join(quadrant)

    @staticmethod
    def _polygon2contour(polygon):

        return [
            (polygon[0], polygon[1]),
            (polygon[0]+polygon[2], polygon[1]),
            (polygon[0]+polygon[2], polygon[1]+polygon[3]),
            (polygon[0], polygon[1]+polygon[3])
        ]

    def _rois_from_img(self, img):
        """
        Return a list of ROIS given a raw input frame.
        """

        rois = []

        if img.shape[2] == 3:
            gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        else:
            gray = img

        binary_img = self._threshold(gray)

        try:
            contours = self._find_contours(binary_img)
        except EthoscopeException:
            raise EthoscopeException("I could not find all ROIs. Please try again or change the lighting conditions.", binary_img)

        for idx, contour in enumerate(contours):
            center = self._find_center(contour)
            quadrant = self._find_quadrant(img.shape, center)
            side = 'right'
            if quadrant[1] == 'l':
                side = 'left'

            roi = ROI(contour, idx=idx+1, side=side)
            rois.append(roi)
            logger.debug("x/200 for roi %d: %d", roi.idx, roi.get_feature_dict()['x'] // 200)


        rois = sorted(rois, key=lambda roi: (roi.get_feature_dict()['x'] // 200, roi.get_feature_dict()['y']))
        for idx, roi in enumerate(rois):
            roi._idx = idx + 1

        return rois

    def _get_roi_score_map(self, img, threshold_value):

        # start a score_map set to 0 everywhere
        score_map = np.zeros(img.shape, dtype=np.uint8)

        # perform binary thresholding
        thresh = cv2.threshold(img, threshold_value, 255, cv2.THRESH_BINARY)[1]

        # cv2.imshow('thresh', thresh)
        # cv2.waitKey(0)

        # morphological operations to denoise
        morph = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, np.ones((5, 5)), iterations=5)

        # pad with only 1 pixel to avoid segmented regions touching the wall
        padded = cv2.copyMakeBorder(morph, 1, 1, 1, 1, cv2.BORDER_CONSTANT)

        # get edges
        edged = cv2.Canny(padded, 50, 100)

        # cv2.imshow('edged', edged)
        # cv2.waitKey(0)

        if CV_VERSION == 3:
            _, contours, _ = cv2.findContours(edged, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        else:
            contours, _ = cv2.findContours(edged, cv2.RETR_EXTERNAL, CHAIN_APPROX_SIMPLE)


        # go through each contour and validate it
        for contour in contours:
            epsilon = 0.001 * cv2.arcLength(contour, True)
            approx_contour = cv2.approxPolyDP(contour, epsilon, True)
            area = cv2.contourArea(approx_contour)

            if self._min_area < area < self._max_area:
                logger.debug(area)
                cv2.drawContours(score_map, [approx_contour], -1, 1, -1)

        return score_map

    def _threshold(self, img):
        r"""
        ROIs are segmented by:
        * BILFILT
        * THRESH LOOP
           1. MORPH X 5
           2. SURROUND BY BLACK TO AVOID WALL EFFECT
           3. CANNY (edge detection)
           4. FIND CONTS
           5. CONT LOOP
           6. APPROX CONTOUR
           7. MINAREARECT
           8. IF WIDTH AND HEIGHT ARE OK
           9. UPDATE SCORE MAP
        """
        image = img.copy()

        # store the addition of all the score_map
        # computed on each iteration
        placeholder = np.zeros(image.shape, dtype=np.uint8)
        blur = cv2.bilateralFilter(img, 10, 150, 150)

        for threshold_value in range(self._min_thresh, self._max_thresh, self._inter_thresh):
            score_map = self._get_roi_score_map(blur, threshold_value)
            placeholder = cv2.add(placeholder, score_map)

        # consider foreground any pixel with at least 10 matches
        binary_image = cv2.threshold(placeholder, 5, 255, cv2.THRESH_BINARY)[1]
        return binary_image



__author__ = 'quentin'

from threading import Thread
import time
import collections

import urllib.request, urllib.error, urllib.parse
import json


class HardwareConnection(Thread):
    def __init__(self, interface_class, *args, **kwargs):
        """
        A class to build a connection to arbitrary hardware.
        It implements an instance of a :class:`~ethoscope.hardware.interfaces.interfaces.BaseInterface` which it uses to send instructions, asynchronously and  on demand.

        :param interface_class: the class to use a an interface to hardware (derives from :class:`~ethoscope.hardware.interfaces.interfaces.BaseInterface`)
        :type interface_class: class
        :param args: list of arguments passed the the hardware interface
        :param kwargs: list of keyword arguments passed the the hardware interface
        """
        self._interface_class = interface_class
        self._interface_args = args
        self._interface_kwargs = kwargs

        self._interface = interface_class(*args, **kwargs)
        self._instructions = collections.deque()
        self._connection_open = True
        super(HardwareConnection, self).__init__()
        self.start()

    def run(self):
        """
        Infinite loop that send instructions to the hardware interface
        Do not call directly, used the ``start()`` method instead.
        """
        while self._connection_open:
            time.sleep(.1)
            while len(self._instructions) > 0 and self._connection_open:
                instruc = self._instructions.popleft()
                ret = self._interface.send(**instruc)

    def send_instruction(self, instruction=None):
        """
        Stage an instruction to be sent to the hardware interface.
        Instructions will be parsed sequentially, but asynchronously from the main thread execution.
        :param instruction: a dictionary of keyword arguments ,matching those of ``interface_class.send()``.
        :type instruction: dict()
        """
        if instruction is None:
            instruction = {}
        if not isinstance(instruction, dict):
            raise Exception("instructions should be dictionaries")
        self._instructions.append(instruction)

    def stop(self, error=None):
        self._connection_open = False
    def __del__(self):
        self.stop()

    def __getstate__(self):
        return {
                "interface_class": self._interface_class,
                "interface_args": self._interface_args,
                "interface_kwargs": self._interface_kwargs}

    def __setstate__(self, state):
        kwargs = state["interface_kwargs"]
        kwargs["do_warm_up"] = False
        self.__init__(state["interface_class"],
                      *state["interface_args"], **kwargs)


class BaseInterface:
    def __init__(self, *args, do_warm_up=True, **kwargs):
        """
        Template class which is an abstract representation of an hardware interface.
        It must define, in :func:`~ethoscope.hardware.interfaces.interfaces.BaseInterface.__init__`,
        how the interface is connected, and in :func:`~ethoscope.hardware.interfaces.interfaces.BaseInterface.send`,
        how information are communicated to the hardware. In addition, derived classes must implement a
        :func:`~ethoscope.hardware.interfaces.interfaces.BaseInterface._warm_up`, method, which defines optionnal instructions
        passed to the hardware upon first connection (that is useful to experimenters that want to manually check their settings).

        :param do_warm_up:
        """
        if do_warm_up:
            self._warm_up()

        super().__init__(*args, **kwargs)

    def _warm_up(self):
        raise NotImplementedError
    def send(self, **kwargs):
        """
        Method to request hardware interface to interact with the physical world.
        :param kwargs: keywords arguments
        """
        raise NotImplementedError


class DefaultInterface(BaseInterface):
    """
    Class that implements a dummy interface that does nothing. This can be used to keep software consistency when
    no hardware is to be used.
    """
    def _warm_up(self):
        pass
    def send(self, **kwargs):
        pass

class ScanException(Exception):
    pass

class EthoscopeSensor(object):
    """
    Class providing access to an ESP32 based WIFI ethoscope sensor
    """

    _sensor_values = {"temperature" : "FLOAT",
                      "humidity": "FLOAT",
                      "light": "INT",
                      "pressure" : "FLOAT"}


    def __init__(self, sensor_url):
        self._sensor_url = sensor_url
        self._last_read = 0
        self._update(True)

    def _get_json_from_url(self, url, timeout=5, post_data=None):
        try:
            if not url.startswith("http://"): url = "http://" + url
            req = urllib.request.Request(url, data=post_data, headers={'Content-Type': 'application/json'})
            f = urllib.request.urlopen(req, timeout=timeout)
            message = f.read()
            if not message:
                # logging.error("URL error whist scanning url: %s. No message back." % self._id_url)
                raise ScanException("No message back")
            try:
                resp = json.loads(message)
                return resp
            except ValueError:
                # logging.error("Could not parse response from %s as JSON object" % self._id_url)
                raise ScanException("Could not parse Json object")

        except urllib.error.HTTPError as e:
            raise ScanException("Error" + str(e.code))
            #return e

        except urllib.error.URLError as e:
            raise ScanException("Error" + str(e.reason))
            #return e

        except Exception as e:
            raise ScanException("Unexpected error" + str(e))

    def _update(self, force=False, freq=5):
        """
        Refresh sensor values
        It is usually a good idea not to interrogate the sensors too often
        to avoid overheating
        freq is the max interval in seconds
        """
        if (time.time() - self._last_read) > freq or force:
            self._sensor_data = self._get_json_from_url(self._sensor_url)
            self._last_read = time.time()

    def read_all(self):
        self._update()
        return tuple( [self._sensor_data[name] for name in self.sensor_properties] )

    @property
    def sensor_properties(self):
        self._update()
        return self._sensor_values.keys()

    @property
    def sensor_types(self):
        self._update()
        return self._sensor_values

    @property
    def temperature(self):
        self._update()
        return self._sensor_data["temperature"]

    @property
    def humidity(self):
        self._update()
        return self._sensor_data["humidity"]

    @property
    def light(self):
        self._update()
        return self._sensor_data["light"]

    @property
    def pressure(self):
        self._update()
        return self._sensor_data["pressure"]




# Standard library imports
import logging
import os
import os.path
#import traceback

# Third party imports
import cv2
import pandas as pd

# Local application imports
# from idoc.decorators import if_record_event, if_not_record_end_event
from idoc.configuration import IDOCConfiguration
from idoc.helpers import get_machine_id, get_machine_name
from idoc.server.core.base import Settings, Status

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

config = IDOCConfiguration()

class CSVResultWriter(Settings, Status):
    r"""
    Save results to csv files styled in the format:
    start_datetime + '_idoc_' + self._machine_id + _TABLENAME .csv
    """

    _quality = cv2.IMWRITE_JPEG_QUALITY

    def __init__(self, *args, nrois=20, max_n_rows_to_insert=10, **kwargs):

        super().__init__(*args, **kwargs)

        self._max_n_rows_to_insert = max_n_rows_to_insert
        updated_tables = ["ROI_%d" % i for i in range(1, nrois+1)]
        updated_tables.append("CONTROLLER_EVENTS")
        self._updated_tables = updated_tables
        self._static_tables = ["METADATA", "ROI_MAP", "VAR_MAP"]

        self._tables = self._updated_tables.extend(self._static_tables)
        self._cache = {table: [] for table in self._updated_tables}
        self._root_dir = config.content["folders"]["results"]["path"]
        self._machine_id = get_machine_id()

        self._var_map_initialised = False
        self._metadata_initialised = False
        self._roi_map_initialised = False
        self.last_t = None

        self.start_datetime = ""
        self._result_dir = ""
        self._output_csv = None
        self._run_id_long = None


    @property
    def info(self):
        return {
            "result_dir": self.result_dir
        }

    @property
    def run_id_long(self):
        self._run_id_long = "%s_%s" % (self.start_datetime, self._machine_id)
        return self._run_id_long


    @property
    def result_dir(self):

        if self.running:
            logger.debug('Output will be saved in %s', self._result_dir)

            self._result_dir = os.path.join(
                 self._root_dir,
                 self._machine_id,
                 get_machine_name(),
                self.start_datetime
            )

        else:
            self._result_dir = ""

        return self._result_dir

    @property
    def img_snapshots(self):
        return os.path.join(
            self.result_dir,
            "IMG_SNAPSHOTS"
        )

    @property
    def snapshots_template(self):
        return os.path.join(
            self.img_snapshots,
            "FRAME_%s.jpg"
        )

    @property
    def output_csv(self):
        self._output_csv = os.path.join(self.result_dir, self.run_id_long + ".csv")


    def set_running(self):

        self.running = True

        try:
            os.makedirs(self.result_dir, exist_ok=False)
        except FileExistsError:
            logger.warning('Output directory already exists. This is weird. Are you passing a custom start_datetime?')
            logger.warning(self.result_dir)



    def initialise_metadata(self, metadata):

        if not self._metadata_initialised:
            os.makedirs(self._result_dir, exist_ok=True)
            metadata.to_csv(
                os.path.join(
                    self._result_dir,
                    self.run_id_long + "_METADATA.csv"
                )
            )

            self._metadata_initialised = True

    def initialise_roi_map(self, rois):

        if not self._roi_map_initialised:

            roi_map = []

            for roi in rois:
                roi_map.append(roi.get_feature_dict())

            roi_map_df = pd.DataFrame(roi_map)
            roi_map_df.to_csv(
                os.path.join(
                    self._result_dir,
                    self.run_id_long + "_ROI_MAP.csv"
                )
            )
            self._roi_map_initialised = True




    def initialise_var_map(self, data_rows):
        """
        Store the name of the columns and their functional type in a separate table
        """

        var_map = []

        for data in data_rows.values():
            try:
                var_map.append({
                    "var_name": data.header_name,
                    "sql_type": data.sql_data_type,
                    "functional_type": data.functional_type
                })
            except AttributeError as error:
                logger.warning(error)
                logger.warning(data)


        var_map_df = pd.DataFrame(var_map)
        # NOTE Refactor code so the makedirs is only run once
        if not os.path.exists(self._result_dir):
            logger.warning(f"Creating {self._result_dir}")
            os.makedirs(self._result_dir, exist_ok=True)


        var_map_df.to_csv(
            os.path.join(
                self._result_dir,
                self.run_id_long  + "_VAR_MAP.csv"
            )
        )

        self._var_map_initialised = True

    def write(self, roi, data_rows):
        """
        An adaptor for the DataPoint instances to be compatible
        with process_row
        """
        if self.running:

            if not self._var_map_initialised:
                self.initialise_var_map(data_rows[0])

            table_name = "ROI_%d" % roi.idx
            for row in data_rows:
                data_points = {**{"id": 0}, **{v.header_name: v for v in row.values()}}
            self.process_row(data_points, table_name)


    def write_frames(self, img, idx):

        if not os.path.isdir(self.img_snapshots):
            os.makedirs(self.img_snapshots)

        cv2.imwrite(self.snapshots_template % str(idx).zfill(7), img, [int(self._quality), 50])


    def process_row(self, data, table_name):
        """
        Store a new data point only if running and not stopped.
        """

        if self.running and not self.stopped:
            data = {k.strip(" "): data[k] for k in data}

            data.update({"t": self.last_t})
            data.update({"elapsed_seconds": self.elapsed_seconds})

            self._cache[table_name].append(data)


            if len(self._cache[table_name]) >= self._max_n_rows_to_insert:
                self.store_and_clear(table_name)

    def get_table_path(self, table_name):
        """
        Generate the absolute path to one of the tables
        generated in the output.
        """
        filename = "%s_%s.csv" % (self.run_id_long, table_name.upper())
        table_path = os.path.join(self._result_dir, filename)
        return table_path

    def store_and_clear(self, table_name):
        """
        Convert the cache list to a DataFrame and append that to HDF5.
        """

        table_path = self.get_table_path(table_name)

        # Generate a pandas dataframe from the cached list
        records = pd.DataFrame.from_records(self._cache[table_name])

        # Append the generated dataframe to the result_file
        if os.path.exists(table_path):
            mode = "a"
            header = False
        else:
            mode = "w"
            header = True

        records.to_csv(table_path, mode=mode, header=header)
        # Once the cache has been written to disk,
        # clear it from RAM
        self._cache[table_name].clear()


    def clear(self):
        """
        Emtpy the cache of all the tables.
        """
        self.stopped = True
        for table in self._cache:
            self.store_and_clear(table)

        self.reset()



import collections
import copy

__author__ = 'quentin'


class DataPoint(collections.OrderedDict):

    def __init__(self, data):
        """
        A container to store variables. It derived from :class:`~collections.OrderedDict`.
        Variables are accessible by header name, which is an individual identifier
        of a variable type (see :class:`~ethoscope.core.variables.BaseIntVariable`):

        >>> from ethoscope.core.variables import DataPoint, XPosVariable, YPosVariable, HeightVariable
        >>> y = YPosVariable(18)
        >>> x = XPosVariable(32)
        >>> data = DataPoint([x,y])
        >>> print data["x"]
        >>> h = HeightVariable(3)
        >>> data.append(h)
        >>> print data


        :param data: a list of data points
        :type data: list(:class:`~ethoscope.core.variables.BaseIntVariable`)
        """
        collections.OrderedDict.__init__(self)
        for i in data:
            self.__setitem__(i.header_name, i)

    def copy(self):
        """
        Deep copy a data point. Copying using the `=` operator will simply create an alias to a `DataPoint`
        object (i.e. allow modification of the original object).

        :return: a copy of this object
        :rtype: :class:`~ethoscope.core.data_point.DataPoint`
        """
        return DataPoint(copy.deepcopy(list(self.values())))

    def append(self, item):
        """
        Add a new variable in the `DataPoint` The order is preserved.

        :param item: A variable to be added.
        :param item: :class:`~ethoscope.core.variables.BaseIntVariable`
        :return:
        """
        self.__setitem__(item.header_name, item)



__author__ = 'antonio'

import logging
import time
import warnings
import cv2
from idoc.server.core.base import Settings, Status
from idoc.debug import IDOCException

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Tell pylint everything here is abstract classes
# pylint: disable=W0223


class BaseCamera(Settings, Status):

    def __init__(self, *args, drop_each=1, max_duration=None, use_wall_clock=True, rotation=0, **kwargs):
        """
        The template class to generate and use video streams.

        :param drop_each: keep only ``1/drop_each``'th frame
        :param max_duration: stop the video stream if ``t > max_duration`` (in seconds).
        :param args: additional arguments
        :param kwargs: additional keyword arguments
        """
        super().__init__(*args, **kwargs)

        self._settings.update({"drop_each": drop_each})
        self.capture = None
        self._max_duration = max_duration
        self._frame_idx = 0
        self._shape = (None, None)
        self._rotation=rotation
        self._use_wall_clock = use_wall_clock
        logger.info("Using wall clock is %r", self._use_wall_clock)

    # TODO Why are 3 params beyond self needed here?
    def __exit__(self): # pylint: disable=unexpected-special-method-signature
        logger.info("Closing camera")
        self.close()

    def __iter__(self):
        """
        Iterate thought consecutive frames of this camera.

        :return: the time (in ms) and a frame (numpy array).
        :rtype: (int, :class:`~numpy.ndarray`)
        """
        at_least_one_frame = False
        while not self.stopped:
            if self.is_last_frame() or not self.is_opened():
                if not at_least_one_frame:
                    raise Exception("Camera could not read the first frame")
                break
            time_s, out = self._next_time_image()
            if out is None:
                break
            t_ms = int(1000 * time_s)
            at_least_one_frame = True

            if (self._frame_idx % self._settings["drop_each"]) == 0:
                logger.debug("Time: %s, Framerate: %s", t_ms, self.framerate)
                yield t_ms, out

            if self._max_duration is not None and t_ms > self._max_duration * 1000:
                break

    def _next_time_image(self):
        timestamp = self._time_stamp()
        image = self._next_image()

        if self._rotation == 0:
            pass
        elif self._rotation == 90:
            image=cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
        elif self._rotation == -90:
            image=cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
        else:
            warnings.warn("Please pass a valid rotation value. Choices are 0, 90 -90")


        self._frame_idx += 1
        return timestamp, image

    def is_last_frame(self):
        raise NotImplementedError

    def _next_image(self):
        raise NotImplementedError

    def _time_stamp(self):
        raise NotImplementedError

    def is_opened(self):
        raise NotImplementedError

    def open(self):
        # super().start()
        raise NotImplementedError

    def close(self):
        super().stop()

    def restart(self):
        """
        Restarts a camera (also resets time).
        :return:
        """
        raise NotImplementedError

class AdaptorCamera(BaseCamera):
    r"""
    Abstract adapter class based on the ethoscope BaseCamera
    providing camera input functionality to IDOC cameras
    """

    framerate = 0
    exposure_time = 0

    def __init__(self, *args, **kwargs):


        super().__init__(*args, **kwargs)
        self.open()
        self._start_time = time.time()
        self._settings.update({
            "video_path": None
        })


    def __str__(self):
        template = '%s @ %s FPS and exposure time %s ms'
        return template % (
            self.__class__.__name__,
            str(self.framerate).zfill(4),
            str(self.exposure_time).zfill(8)
        )

    def open(self):
        super().run()

    def close(self):
        super().stop()

    @staticmethod
    def _validate(img):
        if img is not None:
            if len(img.shape) != 3:
                raise IDOCException("camera output is not 3D")


    @property
    def resolution(self):
        r"""
        Convenience function to return resolution of camera.
        Resolution = (number_horizontal_pixels, number_vertical_pixels)
        """
        return self._settings["resolution"]

    @resolution.setter
    def resolution(self):
        logger.warning("NotImplemented")



# Standard library imports
import datetime
import logging
import os
import random
import warnings

# Third party imports
import bottle

from .configuration import MACHINE_NAME, MACHINE_ID

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class MachineDatetime(datetime.datetime):
    r"""
    A convenient class that expands datetime.datetime
    by having an extra format which matches the one
    IDOC uses to write timestamps in folders and files.
    """

    def machineformat(self):
        year = self.year
        month = self.month
        day = self.day

        hour = self.hour
        minute = self.minute
        second = self.second

        date = "%s-%s-%s" % (str(year).zfill(4), str(month).zfill(2), str(day).zfill(2))
        clock_time = "%s-%s-%s" % (str(hour).zfill(2), str(minute).zfill(2), str(second).zfill(2))
        return "%s_%s" % (date, clock_time)


def hours_minutes_seconds(seconds):
    if seconds is None:
        return None

    return int(seconds // 3600), int((seconds // 60) % 60), int(seconds % 60)

def iso_format(dhm):
    if dhm is None:
        return ""

    hours, minutes, seconds = dhm
    return f'{hours:02d}:{minutes:02d}:{seconds:02d}'


def get_server(port):
    r"""
    This checks if the patch has to be applied or not. We check if bottle has declared cherootserver
    we assume that we are using cherrypy > 9.
    If cherootserver is not available, trick bottle to think cheroot actually is a cherrypy server,
    modifies the server_names allowed in bottle so we use cheroot in background.
    """



    server = "cheroot"
    try:
        from bottle import CherootServer # pylint: disable=import-outside-toplevel

    except ImportError as error:
        server = "cherrypy"

        class CherootServer(bottle.ServerAdapter):

            def __init__(self, *args, **kwargs):

                self._server = None
                super().__init__(*args, **kwargs)

            def run(self, handler): # pragma: no cover
                from cheroot import wsgi # pylint: disable=import-outside-toplevel
                from cheroot.ssl import builtin # pylint: disable=import-outside-toplevel
                self.options['bind_addr'] = (self.host, self.port)
                self.options['wsgi_app'] = handler
                certfile = self.options.pop('certfile', None)
                keyfile = self.options.pop('keyfile', None)
                chainfile = self.options.pop('chainfile', None)
                self._server = wsgi.Server(**self.options)
                if certfile and keyfile:
                    server.ssl_adapter = builtin.BuiltinSSLAdapter(
                        certfile, keyfile, chainfile
                    )

                try:
                    warnings.filterwarnings(
                        action="ignore", message="unclosed",
                        category=ResourceWarning
                    )
                    self._server.start()
                finally:
                    self._server.stop()

        bottle.server_names["cherrypy"] = CherootServer(host='0.0.0.0', port=port)
        logger.info("Cherrypy version is bigger than 9, we have to change to cheroot server")
        # logger.warning(error)

    return server

def get_machine_id():
    r"""
    Reads the machine ID
    This file should be present on any linux installation
    When missing, it is automatically generated by the OS.
    """

    with open(MACHINE_ID, 'r') as filehandle:
        info = filehandle.readline().rstrip()
    return info

def get_commit_version(commit):
    '''
    '''
    commit_date = datetime.datetime.utcfromtimestamp(commit.committed_date)
    commit_date_formatted = commit_date.strftime('%Y-%m-%d %H:%M:%S')

    return {
        "id": str(commit),
        "date": commit_date_formatted
        }


def get_machine_name():
    """
    Reads the machine name
    This file will be present only on a real ethoscope
    When running locally, it will generate a randome name
    """

    if os.path.exists(MACHINE_NAME):
        with open(MACHINE_NAME, 'r') as filehandle:
            info = filehandle.readline().rstrip()
        return info

    else:
        logger.warning(f"Could not find {MACHINE_NAME}. Giving random name")
        return 'LEARNMEMORY_' + str(random.randint(100, 999))



__author__ = 'quentin'

from collections import deque
from math import log10, sqrt, pi
import logging

import cv2
import numpy as np
from scipy import ndimage

from idoc.server.trackers.trackers import BaseTracker, NoPositionError
from idoc.server.core.variables import XPosVariable, YPosVariable, \
    XYDistance, WidthVariable, HeightVariable, PhiVariable
from idoc.server.core.data_point import DataPoint
from idoc.debug import IDOCException

try:
    CV_VERSION = int(cv2.__version__.split(".")[0])
except:
    CV_VERSION = 2


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class ObjectModel():
    """
    A class to model, update and predict foreground object (i.e. tracked animal).
    """
    _sqrt_2_pi = sqrt(2.0 * pi)
    def __init__(self, *args, history_length=1000, **kwargs):
        # FIXME this should be time, not number of points!
        self._features_header = [
            "fg_model_area",
            "fg_model_height",
            #"fg_model_aspect_ratio",
            "fg_model_mean_grey"
        ]

        self._history_length = history_length
        self._ring_buff = np.zeros(
            (self._history_length, len(self._features_header)),
            dtype=np.float32, order="F"
        )

        self._std_buff = np.zeros(
            (self._history_length, len(self._features_header)),
            dtype=np.float32, order="F"
        )

        self._ring_buff_idx = 0
        self._is_ready = False
        self._roi_img_buff = None
        self._mask_img_buff = None
        self._img_buff_shape = np.array([0, 0])

        self._last_updated_time = 0
        # If the model is not updated for this duration, it is reset. Patches #39
        self._max_unupdated_duration = 1 *  60 * 1000.0 #ms

        super().__init__(*args, **kwargs)

    @property
    def is_ready(self):
        return self._is_ready

    @property
    def features_header(self):
        return self._features_header

    def update(self, img, contour, time):
        self._last_updated_time = time
        self._ring_buff[self._ring_buff_idx] = self.compute_features(img, contour)

        self._ring_buff_idx += 1

        if self._ring_buff_idx == self._history_length:
            self._is_ready = True
            self._ring_buff_idx = 0


        return self._ring_buff[self._ring_buff_idx]

    def distance(self, features, time):
        if time - self._last_updated_time > self._max_unupdated_duration:
            logging.warning("FG model not updated for too long. Resetting.")
            self.__init__(history_length=self._history_length)
            return 0

        if not self._is_ready:
            last_row = self._ring_buff_idx + 1
        else:
            last_row = self._history_length

        means = np.mean(self._ring_buff[:last_row], 0)

        np.subtract(self._ring_buff[:last_row], means, self._std_buff[:last_row])
        np.abs(self._std_buff[:last_row], self._std_buff[:last_row])

        stds = np.mean(self._std_buff[:last_row], 0)
        if (stds == 0).any():
            return 0

        a = 1 / (stds * self._sqrt_2_pi)

        b = np.exp(- (features - means) ** 2  / (2 * stds ** 2))

        likelihoods = a * b

        if np.any(likelihoods == 0):
            return 0
        #print features, means
        logls = np.sum(np.log10(likelihoods)) / len(likelihoods)
        return -1.0 * logls


    def compute_features(self, img, contour):
        x, y, w, h = cv2.boundingRect(contour)

        if self._roi_img_buff is None or np.any(self._roi_img_buff.shape < img.shape[0:2]):
            # dynamically reallocate buffer if needed
            self._img_buff_shape[1] = max(self._img_buff_shape[1],w)
            self._img_buff_shape[0] = max(self._img_buff_shape[0], h)

            self._roi_img_buff = np.zeros(self._img_buff_shape, np.uint8)
            self._mask_img_buff = np.zeros_like(self._roi_img_buff)

        sub_mask = self._mask_img_buff[0 : h, 0 : w]

        sub_grey = self._roi_img_buff[0 : h, 0: w]

        cv2.cvtColor(img[y : y + h, x : x + w, :], cv2.COLOR_BGR2GRAY, sub_grey)
        sub_mask.fill(0)

        cv2.drawContours(sub_mask, [contour], -1, 255, -1, offset=(-x,-y))
        mean_col = cv2.mean(sub_grey, sub_mask)[0]


        (_,_) ,(width, height), angle = cv2.minAreaRect(contour)
        width, height = max(width, height), min(width, height)
        ar = ((height+1) / (width+1))

        # TODO speed should use time
        #
        # if len(self.positions) > 2:
        #
        #     pm, pmm = self._positions[-1],self._positions[-2]
        #     xm, xmm = pm["x"], pmm["x"]
        #     ym, ymm = pm["y"], pmm["y"]
        #
        #     instantaneous_speed = abs(xm + 1j*ym - xmm + 1j*ymm)
        # else:
        #     instantaneous_speed = 0
        # if np.isnan(instantaneous_speed):
        #     instantaneous_speed = 0

        features = np.array([
            log10(cv2.contourArea(contour) + 1.0),
            height + 1,
            #sqrt(ar),
            #instantaneous_speed +1.0,
            mean_col +1
            # 1.0
        ])

        return features


class BackgroundModel:
    """
    A class to model background. It uses a dynamic running average and support arbitrary and heterogeneous frame rates
    """
    def __init__(self, *args, max_half_life=500. * 1000, min_half_life=5.* 1000, increment=1.2, **kwargs):
        # the maximal half life of a pixel from background, in seconds
        self._max_half_life = float(max_half_life)
        # the minimal one
        self._min_half_life = float(min_half_life)

        # starts with the fastest learning rate
        self._current_half_life = self._min_half_life

        # FIXME theoretically this should depend on time, not frame index
        self._increment = increment
        # the mean background
        self._bg_mean = None
        # self._bg_sd = None

        self._buff_alpha_matrix = None
        self._buff_invert_alpha_mat = None
        # the time stamp of the frame las used to update
        self.last_t = 0
        super().__init__(*args, **kwargs)

    @property
    def bg_img(self):
        return self._bg_mean

    def increase_learning_rate(self):
        self._current_half_life /= self._increment

    def decrease_learning_rate(self):
        self._current_half_life *= self._increment


    def update(self, img_t, t, fg_mask=None):
        dt = float(t - self.last_t)
        if dt < 0:
            raise NoPositionError("Negative time interval between two consecutive frames")

        # clip the half life to possible value:
        self._current_half_life = np.clip(
            self._current_half_life,
            self._min_half_life,
            self._max_half_life
        )

        # ensure preallocated buffers exist. otherwise, initialise them
        if self._bg_mean is None:
            self._bg_mean = img_t.astype(np.float32)
            # self._bg_sd = np.zeros_like(img_t)
            # self._bg_sd.fill(128)

        if self._buff_alpha_matrix is None:
            self._buff_alpha_matrix = np.ones_like(img_t, dtype=np.float32)

        # the learning rate, alpha, is an exponential function of half life
        # it correspond to how much the present frame should account for the background

        lam = np.log(2) / self._current_half_life
        # how much the current frame should be accounted for
        alpha = 1 - np.exp(-lam * dt)

        # set-p a matrix of learning rate. it is 0 where foreground map is true
        self._buff_alpha_matrix.fill(alpha)
        if fg_mask is not None:
            cv2.dilate(fg_mask, None, fg_mask)
            cv2.subtract(self._buff_alpha_matrix, self._buff_alpha_matrix, self._buff_alpha_matrix, mask=fg_mask)


        if self._buff_invert_alpha_mat is None:
            self._buff_invert_alpha_mat = 1 - self._buff_alpha_matrix
        else:
            np.subtract(1, self._buff_alpha_matrix, self._buff_invert_alpha_mat)

        np.multiply(self._buff_alpha_matrix, img_t, self._buff_alpha_matrix)
        np.multiply(self._buff_invert_alpha_mat, self._bg_mean, self._buff_invert_alpha_mat)
        np.add(self._buff_alpha_matrix, self._buff_invert_alpha_mat, self._bg_mean)

        self.last_t = t


class AdaptiveBGModel(BaseTracker):
    _description = {"overview": "The default tracker for fruit flies. One animal per ROI.",
                    "arguments": []}

    fg_model = ObjectModel()

    def __init__(self, *args, **kwargs):
        """
        An adaptive background subtraction model to find position of one animal in one roi.

        TODO more description here
        :param roi:
        :param data:
        :return:
        """
        self._previous_shape = None
        self._object_expected_size = 0.05 # proportion of the roi main axis
        self._max_area = (5 * self._object_expected_size) ** 2

        self._smooth_mode = deque()
        self._smooth_mode_tstamp = deque()
        self._smooth_mode_window_dt = 30 * 1000 #miliseconds


        self._max_m_log_lik = 5.5
        self._buff_grey = None
        self._buff_object = None
        self._buff_object_old = None
        self._buff_grey_blurred = None
        self._buff_fg = None
        self._buff_convolved_mask = None
        self._buff_fg_backup = None
        self._buff_fg_diff = None
        self._old_sum_fg = 0
        self._bg_model = BackgroundModel()
        super().__init__(*args, *kwargs)


    def _pre_process_input_minimal(self, img, mask, t, darker_fg=True):
        """

        Return a standardised input frame so the tracking functionality
        has an easier job to be done

        :param t: time in ms
        :type t: int
        :param img: the whole frame with 3 channels
        :type img: :class:`~numpy.ndarray`
        :return: A standardised gray 2D array image
        :rtype: np.array
        """


        blur_rad = int(self._object_expected_size * np.max(img.shape) / 2.0)

        if blur_rad % 2 == 0:
            blur_rad += 1

        if self._buff_grey is None:
            self._buff_grey = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            if mask is None:
                mask = np.ones_like(self._buff_grey) * 255

        cv2.cvtColor(img, cv2.COLOR_BGR2GRAY, self._buff_grey)

        cv2.GaussianBlur(self._buff_grey, (blur_rad, blur_rad), 1.2, self._buff_grey)
        if darker_fg:
            cv2.subtract(255, self._buff_grey, self._buff_grey)

        mean = cv2.mean(self._buff_grey, mask)

        scale = 128. / mean[0]

        cv2.multiply(self._buff_grey, scale, dst=self._buff_grey)


        if mask is not None:
            cv2.bitwise_and(self._buff_grey, mask, self._buff_grey)
            return self._buff_grey

    # def _pre_process_input(self, img, mask, t):
    #     """
    #     Locate the animal in a image, at a given time.

    #     :param t: time in ms
    #     :type t: int
    #     :param img: the whole frame with 3 channels
    #     :type img: :class:`~numpy.ndarray`
    #     :return: The position of the animal at time ``t``
    #     :rtype: :class:`~ethoscope.core.data_point.DataPoint`
    #     """

    #     if len(img.shape) != 3:
    #         raise IDOCException("tracker.track: input image is not 3D")

    #     blur_rad = int(self._object_expected_size * np.max(img.shape) * 2.0)
    #     if blur_rad % 2 == 0:
    #         blur_rad += 1


    #     if self._buff_grey is None:
    #         self._buff_grey = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    #         self._buff_grey_blurred = np.empty_like(self._buff_grey)
    #         # self._buff_grey_blurred = np.empty_like(self._buff_grey)
    #         if mask is None:
    #             mask = np.ones_like(self._buff_grey) * 255

    #         mask_conv = cv2.blur(mask, (blur_rad, blur_rad))

    #         self._buff_convolved_mask = (1/255.0 *  mask_conv.astype(np.float32))

    #     cv2.cvtColor(img, cv2.COLOR_BGR2GRAY, self._buff_grey)

    #     hist = cv2.calcHist([self._buff_grey], [0], None, [256], [0, 255]).ravel()
    #     hist = np.convolve(hist, [1] * 3)
    #     mode = np.argmax(hist)

    #     self._smooth_mode.append(mode)
    #     self._smooth_mode_tstamp.append(t)

    #     if len(self._smooth_mode_tstamp) > 2 and self._smooth_mode_tstamp[-1] - self._smooth_mode_tstamp[0] > self._smooth_mode_window_dt:
    #         self._smooth_mode.popleft()
    #         self._smooth_mode_tstamp.popleft()


    #     mode = np.mean(list(self._smooth_mode))
    #     scale = 128. / mode

    #     # cv2.GaussianBlur(self._buff_grey,(5,5), 1.5,self._buff_grey)

    #     cv2.multiply(self._buff_grey, scale, dst=self._buff_grey)

    #     cv2.bitwise_and(self._buff_grey, mask, self._buff_grey)

    #     cv2.blur(self._buff_grey, (blur_rad, blur_rad), self._buff_grey_blurred)
    #     # FIXME could be optimised
    #     self._buff_grey_blurred = (self._buff_grey_blurred / self._buff_convolved_mask).astype(np.uint8)


    #     cv2.absdiff(self._buff_grey, self._buff_grey_blurred, self._buff_grey)

    #     if mask is not None:
    #         cv2.bitwise_and(self._buff_grey, mask, self._buff_grey)
    #         return self._buff_grey


    def _find_position(self, img, mask, t):
        """
        Locate the animal in a image, at a given time

        1 - Preprocess the input frame to have standard properties
        2 - Actually find the position of the animal

        :param t: time in ms
        :type t: int
        :param img: the whole frame with 3 channels
        :type img: :class:`~numpy.ndarray`
        :return: The position of the animal at time ``t``
        :rtype: :class:`~ethoscope.core.data_point.DataPoint`
        """

        if len(img.shape) != 3:
            raise IDOCException("tracker.track: input image is not 3D")


        grey = self._pre_process_input_minimal(img, mask, t)
        # grey = self._pre_process_input(img, mask, t)
        try:
            return self._track(img, grey, mask, t)
        except NoPositionError:
            self._bg_model.update(grey, t)
            raise NoPositionError


    def _track(self, img, grey, mask, t):
        """
        :param img: The 3D image output from roi.apply
        i.e. the original frame produced by the camera and then cropped
        to encompass the ROI only.
        :type img: np.array

        :param img: A 2D copy of img with standard properties
        as produced by the preprocessing functionality.
        :type img: np.array
        """

        if len(img.shape) != 3:
            raise IDOCException("tracker._track: Input img is not 3D")
        if len(grey.shape) != 2:
            raise IDOCException("tracker._track: Input grey is not 2D")

        if self._bg_model.bg_img is None:
            self._buff_fg = np.empty_like(grey)
            self._buff_object = np.empty_like(grey)
            self._buff_fg_backup = np.empty_like(grey)
            self._old_pos = 0.0 +0.0j
            raise NoPositionError


        # regeneration of the buffered foreground
        # as the difference between the model of the background
        # and the new grey image
        # differences < 20 are ignored i.e. only keep changes > 20
        bg = self._bg_model.bg_img.astype(np.uint8)
        cv2.subtract(grey, bg, self._buff_fg)
        cv2.threshold(self._buff_fg, 20, 255, cv2.THRESH_TOZERO, dst=self._buff_fg)

        self._buff_fg_backup = np.copy(self._buff_fg)

        n_fg_pix = np.count_nonzero(self._buff_fg)
        prop_fg_pix = n_fg_pix / (1.0 * grey.shape[0] * grey.shape[1])
        is_ambiguous = False

        if  prop_fg_pix > self._max_area:
            self._bg_model.increase_learning_rate()
            raise NoPositionError

        if  prop_fg_pix == 0:
            self._bg_model.increase_learning_rate()
            raise NoPositionError

        if CV_VERSION == 3:
            _, contours, _ = cv2.findContours(self._buff_fg, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        else:
            contours, _ = cv2.findContours(self._buff_fg, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        contours = [cv2.approxPolyDP(c, 1.2, True) for c in contours]

        if len(contours) == 0:
            self._bg_model.increase_learning_rate()
            raise NoPositionError

        elif len(contours) > 1:
            if not self.fg_model.is_ready:
                raise NoPositionError
            hulls = contours

            hulls = [h for h in hulls if h.shape[0] >= 3]

            if len(hulls) < 1:
                raise NoPositionError

            elif len(hulls) > 1:
                is_ambiguous = True
            cluster_features = [self.fg_model.compute_features(img, h) for h in hulls]
            all_distances = [self.fg_model.distance(cf, t) for cf in cluster_features]
            good_clust = np.argmin(all_distances)

            hull = hulls[good_clust]
            distance = all_distances[good_clust]
        else:
            hull = contours[0]
            if hull.shape[0] < 3:
                self._bg_model.increase_learning_rate()
                raise NoPositionError

            features = self.fg_model.compute_features(img, hull)
            distance = self.fg_model.distance(features, t)

        if distance > self._max_m_log_lik:
            self._bg_model.increase_learning_rate()
            raise NoPositionError

        (x, y), (w, h), angle = cv2.minAreaRect(hull)

        if w < h:
            angle -= 90
            w, h = h, w
        angle = angle % 180

        h_im = min(grey.shape)
        w_im = max(grey.shape)
        max_h = 2 * h_im
        if w > max_h or h > max_h:
            raise NoPositionError

        cv2.ellipse(self._buff_fg, ((x, y), (int(w*1.5), int(h*1.5)), angle), 255, -1)

        #TODO center mass just on the ellipse area
        cv2.bitwise_and(self._buff_fg_backup, self._buff_fg, self._buff_fg_backup)

        y, x = ndimage.center_of_mass(self._buff_fg_backup)

        pos = x + 1.0j*y
        pos /= w_im

        xy_dist = round(log10(1./float(w_im) + abs(pos - self._old_pos))*1000)

        self._old_pos = pos

        if mask is not None:
            cv2.bitwise_and(self._buff_fg, mask, self._buff_fg)

        if is_ambiguous:
            self._bg_model.increase_learning_rate()
            self._bg_model.update(grey, t)
        else:
            self._bg_model.decrease_learning_rate()
            self._bg_model.update(grey, t, self._buff_fg)

        self.fg_model.update(img, hull, t)
        x_var = XPosVariable(int(round(x)))
        y_var = YPosVariable(int(round(y)))
        distance = XYDistance(int(xy_dist))
        w_var = WidthVariable(int(round(w)))
        h_var = HeightVariable(int(round(h)))
        phi_var = PhiVariable(int(round(angle)))

        out = DataPoint([
            x_var, y_var,
            w_var, h_var,
            phi_var,
            distance,
        ])

        self._previous_shape = np.copy(hull)
        return [out]



import datetime
import logging
import os.path
from shutil import copyfile
import time
import traceback

import pandas as pd

from idoc.server.hardware.interfaces.interfaces import DefaultInterface
from idoc.server.programmers.programmers import Programmer
from idoc.server.core.base import Base
from idoc.server.controllers.threads import DefaultArduinoThread, WaveArduinoThread
from idoc.server.controllers.threads import DefaultDummyThread, WaveDummyThread
from idoc.debug import IDOCException
from idoc.configuration import IDOCConfiguration

config = IDOCConfiguration()

# Tell pylint everything here is abstract classes
# pylint: disable=undefined-loop-variable

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class Controller(DefaultInterface, Base):
    r"""
    Interface between a programmer, its threads and the control_thread.
    Reports the state of the controller functionality.
    It tipically drives an Arduino board in the real world.
    """

    _programmerClass = Programmer
    valid_actions = ["start", "stop", "reset"]
    _always_on_hardware = ["TARGETS", "MAIN_VALVE", "VACUUM", "IRLED"]
    _threads = {
        "ArduinoDummy": {"DEFAULT": DefaultDummyThread, "WAVE": WaveDummyThread},
        "ArduinoBoard": {"DEFAULT": DefaultArduinoThread, "WAVE": WaveArduinoThread},
        "ArduinoMegaBoard": {"DEFAULT": DefaultArduinoThread, "WAVE": WaveArduinoThread}
    }


    def __init__(
            self, mapping_path, paradigm_path, result_writer, board_class, arduino_port, *args,
            sampling_rate=10.0, **kwargs):
        """
        :param mapping_path: Path to a .csv showing which hardware is phisically wired to which pin.
        :type mapping_path: str

        :param paradigm_path: Path to .csv listing a sequence of pin trigger.
        Read and processed by the programmer class.
        :type paradigm_path: str

        :param args: Extra positional arguments for the threading.Thread constructor.

        :param board_class: # TODO
        :type board: str

        :param arduino_port: Path to device file representing the board in the OS.
        :type arduino_port: str

        :param sampling_rate: Frequency in Hertz with which the pins are updated.
        :type sampling_rate: float

        :param kwargs: Extra keyword arguments for the threading.Thread constructor.
        """

        DefaultInterface.__init__(self)

        self._use_wall_clock = kwargs.pop("use_wall_clock")
        super().__init__(*args, **kwargs)

        self._mapping_path = mapping_path or config.content["controller"]["mapping_path"]
        self._loaded = False
        self._locked = False
        self._paradigm = []

        # Build a dictionary mapping hardware to pin number
        self._mapping = self._load_mapping(self._mapping_path)
        #Store the state of each pin of the controller
        self._pin_state = {hardware: 0 for hardware in self._mapping}

        # Board instance compatible with the threads classes
        self._board_class = board_class
        self._result_writer = result_writer
        self._sampling_rate = sampling_rate
        self.adaptation_offset = 0
        self._arduino_port = arduino_port
        self._board = self._board_class(self._arduino_port)
        self._pins = {}
        self._pin_defs = {}

        # Programmer instance that processes user input
        # into a 'program', an organized sequence of events
        # that turn on and off the pins in the board

        self._submodules['programmer'] = self._programmerClass()

        self.paradigm_path = paradigm_path or config.content["controller"]["paradigm_path"]

        self._settings['programmer'] = self._submodules['programmer'].settings
        self._progress = 0
        self._last_t = 0



    @property
    def last_t(self):
        return self._last_t

    @last_t.setter
    def last_t(self, last_t):
        self._last_t = last_t

    def toggle(self, hardware, value):

        try:
            for thread in self._paradigm:
                if thread.hardware == hardware:

                    if value == 0:
                        logging.info("Toggling %s --> off", hardware)
                        thread.turn_off()
                    else:
                        logging.info("Toggling %s --> %s", hardware, str(value))
                        thread.turn_on(value)

                    return {"status": "success"}

            logger.warning("Hardware %s not found in loaded thread paradigm.")
            return None

        except Exception as error:
            logger.warning(error)
            return

    @property
    def programmer(self):
        return self._submodules["programmer"]


    @property
    def duration(self):
        """
        Return maximum value of end stored in the table
        used to generate the paradigm
        This is considered as the duration of the paradigm
        The value is returned in seconds, assumming the user input was in seconds
        """
        ends = [row["end"] for row in self.programmer.table]
        try:
            duration = max(ends)
            return duration

        except ValueError:
            return None


    @property
    def progress(self):
        """
        Return a float from 0 to 100 indicating how many of the
        self.paradigm_duration seconds have passed since
        the controller started running
        """

        if self._time_zero is None:
            self._progress = 0
            return self._progress

        elif self.stopped:
            return self._progress

        else:
            diff = max((datetime.datetime.now() - self._time_zero).total_seconds(), 0)

        try:
            progress = diff / self.duration
        except TypeError: #programmer.duration is not defined
            progress = 0

        self._progress = progress * 100

        return self._progress

    @property
    def mapping(self):
        return self._mapping

    @mapping.setter
    def mapping(self, mapping_path):
        self._mapping = self._load_mapping(mapping_path)

    @property
    def info(self):
        return self._info

    @info.getter
    def info(self):
        r"""
        Add to the default info dictionary (status and time_zero)
        the following controller specific properties
        """
        # logging.debug('Requesting Controller.info')
        self._info.update({
            "status": self.status,
            "paradigms": self.programmer.list(),
            # "paradigm_path": self.programmer.paradigm_path,
            "pin_state": self.pin_state,
            "mapping": self.mapping,
            "progress": self.progress,
            "board": self._board.__class__.__name__,
            "duration": self.duration
        })
        # logging.debug('Controller.info OK')

        return self._info


    def populate_pins(self, table):
        """
        Take a table produced by programmer._load_table
        (a list of dicts with keys equal to the arguments
        of the thread classes init functions)
        Read which mode is used in each pin and get them from the
        board with the right mode.
        Save these pins under the dictionary self._pins with keys
        equal to the name of the hardware connected to that pin
        """

        seen_hardware = []
        # TODO
        # This code will set the pins wrong
        # if a new paradigm with the same hardware in different mode is uploaded

        print("Loading paradigm: ")
        print(pd.DataFrame.from_records(table))

        for row in table:
            hardware = row['hardware']
            pin_number = self._mapping[hardware]
            mode = row['mode']
            pin_definition = '%s:%s:%s' % ('d', pin_number, mode)
            if hardware in seen_hardware:
                pass
            elif pin_definition in self._pin_defs:
                self._pins[hardware] = self._pin_defs[pin_definition]
                seen_hardware.append(hardware)
            else:
                logger.debug(f"Getting pin {hardware} - {pin_definition}")
                pin = self._board.get_pin(pin_definition)
                self._pins[hardware] = pin
                self._pin_defs[pin_definition] = pin
                seen_hardware.append(hardware)

    @property
    def paradigm_path(self):
        return self._programmer.paradigm_path

    @paradigm_path.setter
    def paradigm_path(self, paradigm_path):

        # don't load again if the program has the same filename
        # this is done to avoid a superfluous update resetting the self.paradigm object
        if paradigm_path != self.programmer._settings["paradigm_path"]:

            # if there already is an initialized board, close it
            # so we can open it again
            # a fresh board is always needed because otherwise a
            # pyfirmata.pyfirmata.PinAlreadyTakenError is raised
            # when the Thread classes attempt to retrieve their pin
            self._loaded = False
            logger.info("Initalizing %s", self._board_class.__name__)
            try:
                self.programmer.paradigm_path = paradigm_path
            except IDOCException as error:
                logger.warning(error)
                self._paradigm.clear()
                return

            self.populate_pins(self.programmer.table)
            self._prepare()


    @property
    def pin_state(self):
        r"""
        A getter so more high level classes can access _pin_state
        without being able to write (set) to it
        """
        return self._pin_state

    def list(self):
        return self.programmer.list()

    def _load_mapping(self, mapping_path):
        r"""
        Read a mapping.csv file and return a dictionary of key-value pairs
        where keys are hardware names and values the pin number they are
        physically wired to.
        """
        mapping_df = pd.read_csv(mapping_path, comment='#')
        required_columns = ['hardware', 'pin_number']

        if not all([e in mapping_df.columns for e in required_columns]):
            logger.warning("Provided mapping is not valid")
            logger.warning("mapping_path: %s", mapping_path)
            logger.warning("Required columns %s", ' '.join(required_columns))
            logger.warning("Provided columns %s", ' '.join(mapping_df.columns))
            return

        mapping = {}
        for row_tuple in mapping_df.iterrows():
            row = row_tuple[1].to_dict()
            mapping[row["hardware"]] = row["pin_number"]

        # TODO additional processing steps to guarantee standard mapping format
        return mapping

    @property
    def columns(self):
        """
        Returns a list with the hardware in use
        Useful to prepare the header of the output .csv.
        """
        return list(self._mapping.keys())

    def load_paradigm(self, paradigm_path):
        if paradigm_path is not None:
            logger.info('Loading paradigm: %s', paradigm_path)
            self.paradigm_path = paradigm_path

    @property
    def ready(self):
        return self.programmer.paradigm_path is not None

    @ready.setter
    def ready(self, value): # pylint: disable=unused-argument
        return None

    @property
    def wait(self):
        return max(0, self.adaptation_offset - self.init_elapsed_seconds)

    @property
    def adaptation_left(self):
        return max(0, self.adaptation_offset - self.wait)

    def run_minimal(self):
        """
        Start executing the threads whose hardware matches the names listed in
        self._always_on_hardware.
        """

        for hardware in self._always_on_hardware:
            for thread in self._paradigm:
                if thread.hardware == hardware:
                    logger.info('Turning on %s', thread.hardware)
                    thread.pin.write(1)
                    # TODO what happens with the time_zero?
        time.sleep(2)

    def tick(self, last_t):
        """
        Update the time in the threads
        """

        self.last_t = last_t
        if self._loaded:
            if self.running:
                for thread in self._paradigm:
                    thread.last_t = self.last_t

    def _prepare(self):
        """
        Given a paradigm_table, parse it
        to produce a paradigm by storing instances of
        ControllerThread in the self._paradigm list.
        A paradigm is a list of ControllerThread instances
        """

        if self._locked:
            return

        self._paradigm.clear()
        logger.debug("Clearing paradigm")

        for i, row in enumerate(self.programmer.table):

            thread, kwargs = self.programmer.parse(row, i)
            kwargs["result_writer"] = self._result_writer
            kwargs["sampling_rate"] = self._sampling_rate
            kwargs["pin"] = self._pins[kwargs['hardware']]

            ThreadClass = self._threads[self._board_class.__name__][thread]

            logger.debug("Selecting %s", ThreadClass.__name__)
            thread = ThreadClass(
                **kwargs, pin_state=self._pin_state, use_wall_clock=self._use_wall_clock
            )

            self._paradigm.append(thread)

        barriers = self.programmer.make_barriers(self.programmer.table)
        self._paradigm = self.programmer.add_barriers(self._paradigm, barriers)
        self._loaded = True


    def lock(self):
        self.programmer.lock()
        self._locked = True


    def copy_paradigm(self):
        """
        Copy the passed paradigm to the result_dir
        for the purpose of documenting the experiment.
        """
        src =  self.programmer.paradigm_path
        filename = os.path.basename(src)

        dst = self._result_writer.result_dir
        copyfile(src, os.path.join(dst, filename))


    def run(self):
        r"""
        Start executing the threads in self._submodules['programmer'].paradigm.
        """

        logger.debug("Controller has elapsed %3.f seconds since init", self.init_elapsed_seconds)
        
        print(f"{self} will wait for {self.wait} seconds")

        while self.wait > 0:
            self._end_event.wait(0.1)

        self.ready = True
        self.lock()
        super().run()
        self.copy_paradigm()

        for thread in self._paradigm:
            try:
                thread.time_zero = self.time_zero

                # catch if the thread is already started because it is listed on the always_on_hardware
                if thread.is_alive() and thread.hardware in self._always_on_hardware:
                    pass
                else:
                    thread.start()

            except Exception as error:
                logger.info(thread)
                logger.error(error)
                logger.error(traceback.print_exc())
                self.reset()

    def stop(self):

        super().stop()

        for thread in self._paradigm:
            thread.stop()

        for thread in self._paradigm:
            if thread.is_alive():
                thread.join()
                logger.debug("Joined %s", thread.name)

        logger.debug("Done joining!")

    # TODO Implement a warmup routine so th user can check the hardware
    def warm_up(self):
        pass


if __name__ == "__main__":

    from idoc.server.io.result_writer import CSVResultWriter

    controller = Controller(
        mapping_path=config.content["controller"]["mapping_path"],
        paradigm_path=config.content["controller"]["paradigm_path"],
        result_writer=CSVResultWriter(
            start_datetime="2000-01-01_00-00-00",
            max_n_rows_to_insert=2
        ),
        board_class=config.content['default_class']["board"],
        arduino_port=config.content["controller"]["arduino_port"]
    )

    controller.start()
    controller.join()



'''
any new class added here need to be added to web_utils/control_thread.py too
'''

__author__ = 'quentin'


from ethoscope.stimulators.stimulators import BaseStimulator, HasInteractedVariable

from ethoscope.hardware.interfaces.interfaces import  DefaultInterface
from ethoscope.hardware.interfaces.sleep_depriver_interface import SleepDepriverInterface, SleepDepriverInterfaceCR
from ethoscope.hardware.interfaces.optomotor import OptoMotor


import random
import time

class IsMovingStimulator(BaseStimulator):
    _HardwareInterfaceClass = DefaultInterface

    def __init__(self, hardware_connection=None, velocity_correction_coef=3.0e-3, date_range = "", **kwargs):
        """
        class implementing an stimulator that decides whether an animal has moved though does nothing   accordingly.
        :param hardware_connection: a default hardware interface object
        :param velocity_correction_coef: the correction coeeficient for computing velocity at various fps. Emirically defined. When greater than one, the animal is moving
        :type velocity_correction_coef: float
        """
        self._velocity_correction_coef = velocity_correction_coef
        self._last_active = 0
        super(IsMovingStimulator, self).__init__(hardware_connection, date_range)


    def _has_moved(self):

        positions = self._tracker.positions

        if len(positions ) <2 :
            return False


        if len(positions[-1]) != 1:
            raise Exception("This stimulator can only work with a single animal per ROI")
        tail_m = positions[-1][0]

        times = self._tracker.times
        last_time_for_position = times[-1]
        last_time = self._tracker.last_time_point

        # we assume no movement if the animal was not spotted
        if last_time != last_time_for_position:
            return False

        dt_s = abs(times[-1] - times[-2]) / 1000.0
        dist = 10.0 ** (tail_m["xy_dist_log10x1000"]/1000.0)
        velocity = dist / dt_s

        velocity_corrected = velocity * dt_s / self._velocity_correction_coef

        if velocity_corrected > 1.0:
            return True
        return False

    def _decide(self):

        has_moved = self._has_moved()

        t = self._tracker.times
        if  has_moved:# or xor_diff > self._xor_speed_threshold :
            self._last_active = t[-1]
            return HasInteractedVariable(False), {}
        return HasInteractedVariable(True), {}

class SleepDepStimulator(IsMovingStimulator):
    _description = {"overview": "A stimulator to sleep deprive an animal using servo motor. See http://todo/fixme.html",
                    "arguments": [
                                    {"type": "number", "min": 0.0, "max": 1.0, "step":0.0001, "name": "velocity_correction_coef", "description": "Velocity correction coef","default":3.0e-3},
                                    {"type": "number", "min": 1, "max": 3600*12, "step":1, "name": "min_inactive_time", "description": "The minimal time after which an inactive animal is awaken","default":120},
                                    {"type": "date_range", "name": "date_range",
                                     "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                     "default": ""},
                                   ]}

    _HardwareInterfaceClass = SleepDepriverInterface
    _roi_to_channel = {
            1:1,  3:2,  5:3,  7:4,  9:5,
            12:6, 14:7, 16:8, 18:9, 20:10
        }
    def __init__(self,
                 hardware_connection,
                 velocity_correction_coef=3.0e-3,
                 min_inactive_time=120,  #s
                 date_range=""
                 ):
        """
        A stimulator to control a sleep depriver module.

        :param hardware_connection: the sleep depriver module hardware interface
        :type hardware_connection: :class:`~ethoscope.hardware.interfaces.sleep_depriver_interface.SleepDepriverInterface`
        :param velocity_correction_coef:
        :type velocity_correction_coef: float
        :param min_inactive_time: the minimal time without motion after which an animal should be disturbed (in seconds)
        :type min_inactive_time: float
        :return:
        """

        self._inactivity_time_threshold_ms = min_inactive_time *1000 #so we use ms internally
        self._t0 = None
        
        super(SleepDepStimulator, self).__init__(hardware_connection, velocity_correction_coef, date_range=date_range)



    def _decide(self):
        roi_id= self._tracker._roi.idx
        now =  self._tracker.last_time_point

        try:
            channel = self._roi_to_channel[roi_id]
        except KeyError:
            return HasInteractedVariable(False), {}

        has_moved = self._has_moved()


        if self._t0 is None:
            self._t0 = now

        if not has_moved:
            if float(now - self._t0) > self._inactivity_time_threshold_ms:
                self._t0 = None
                return HasInteractedVariable(True), {"channel":channel}
        else:
            self._t0 = now
        return HasInteractedVariable(False), {}


class SleepDepStimulatorCR(SleepDepStimulator):
    _description = {"overview": "A stimulator to sleep deprive an animal using servo motor in Continous Rotation mode. See http://todo/fixme.html",
                    "arguments": [
                        {"type": "number", "min": 0.0, "max": 1.0, "step": 0.0001, "name": "velocity_correction_coef", "description": "Velocity correction coef", "default": 3.0e-3},
                                    {"type": "number", "min": 1, "max": 3600*12, "step":1, "name": "min_inactive_time", "description": "The minimal time after which an inactive animal is awaken","default":120},
                                    {"type": "date_range", "name": "date_range",
                                     "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                     "default": ""}
                                   ]}

    _HardwareInterfaceClass = SleepDepriverInterfaceCR
    _roi_to_channel = {
            1:1,  3:2,  5:3,  7:4,  9:5,
            12:6, 14:7, 16:8, 18:9, 20:10
        }
    def __init__(self,
                 hardware_connection,
                 velocity_correction_coef=3.0e-3,
                 min_inactive_time=120,  #s
                 date_range=""
                 ):
        """
        A stimulator to control a sleep depriver module.

        :param hardware_connection: the sleep depriver module hardware interface
        :type hardware_connection: :class:`~ethoscope.hardware.interfaces.sleep_depriver_interface.SleepDepriverInterface`
        :param velocity_correction_coef:
        :type velocity_correction_coef: float
        :param min_inactive_time: the minimal time without motion after which an animal should be disturbed (in seconds)
        :type min_inactive_time: float
        :return:
        """

        self._inactivity_time_threshold_ms = min_inactive_time *1000 #so we use ms internally
        self._t0 = None
        
        super(SleepDepStimulator, self).__init__(hardware_connection, velocity_correction_coef, date_range=date_range)



class OptomotorSleepDepriver(SleepDepStimulator):
    _description = {"overview": "A stimulator to sleep deprive an animal using gear motors. See https://github.com/gilestrolab/ethoscope_hardware/tree/master/modules/gear_motor_sleep_depriver",
                    "arguments": [
                        {"type": "number", "min": 0.0, "max": 1.0, "step": 0.0001, "name": "velocity_correction_coef", "description": "Velocity correction coef", "default": 3.0e-3},
                                    {"type": "number", "min": 1, "max": 3600*12, "step":1, "name": "min_inactive_time", "description": "The minimal time after which an inactive animal is awaken(s)","default":120},
                                    {"type": "number", "min": 500, "max": 10000 , "step": 50, "name": "pulse_duration", "description": "For how long to deliver the stimulus(ms)", "default": 1000},
                                    {"type": "number", "min": 0, "max": 3, "step": 1, "name": "stimulus_type",  "description": "1 = opto, 2= moto", "default": 2},
                                    {"type": "date_range", "name": "date_range",
                                     "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                     "default": ""}
                                   ]}

    _HardwareInterfaceClass = OptoMotor
    _roi_to_channel_opto = {1:1, 3:3, 5:5, 7:7, 9:9,
                            12:23, 14:21,16:19, 18:17, 20:15}
    _roi_to_channel_moto = {1:0, 3:2, 5:4, 7:6, 9:8,
                            12:22, 14:20, 16:18, 18:16, 20:14}


    def __init__(self,
                 hardware_connection,
                 velocity_correction_coef=3.0e-3,
                 min_inactive_time=120,  # s
                 pulse_duration = 1000,  #ms
                 stimulus_type = 2,  # 1 = opto, 2= moto, 3 = both
                 date_range=""
                 ):


        self._t0 = None

        # the inactive time depends on the chanel here
        super(OptomotorSleepDepriver, self).__init__(hardware_connection, velocity_correction_coef, min_inactive_time, date_range)



        if stimulus_type == 2:
            self._roi_to_channel = self._roi_to_channel_moto
        elif stimulus_type == 1:
            self._roi_to_channel = self._roi_to_channel_opto

        self._pulse_duration= pulse_duration

    def _decide(self):
        out, dic = super(OptomotorSleepDepriver, self)._decide()
        dic["duration"] = self._pulse_duration
        return out,dic



class ExperimentalSleepDepStimulator(SleepDepStimulator):
    _description = {"overview": "A stimulator to sleep deprive an animal using servo motor. See http://todo/fixme.html",
                    "arguments": [
                        {"type": "number", "min": 0.0, "max": 1.0, "step": 0.0001, "name": "velocity_correction_coef", "description": "Velocity correction coef", "default": 3.0e-3},
                                    {"type": "date_range", "name": "date_range",
                                     "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                     "default": ""}
                                   ]}

    _HardwareInterfaceClass = SleepDepriverInterface
    _roi_to_channel = {
            1:1,  3:2,  5:3,  7:4,  9:5,
            12:6, 14:7, 16:8, 18:9, 20:10
        }

    def __init__(self,
                 hardware_connection,
                 velocity_correction_coef=3.0e-3,
                 date_range=""
                 ):
        """
        A stimulator to control a sleep depriver module.
        This is an experimental version where each channel has a different inactivity_time_threshold.

        :param hardware_connection: the sleep depriver module hardware interface
        :type hardware_connection: :class:`~ethoscope.hardawre.interfaces.sleep_depriver_interface.SleepDepriverInterface`
        :param velocity_correction_coef:
        :type velocity_correction_coef: float
        :return:
        """

        self._t0 = None

        
        # the inactive time depends on the chanel here
        super(ExperimentalSleepDepStimulator, self).__init__(hardware_connection, velocity_correction_coef, 0, date_range)
        self._inactivity_time_threshold_ms = None

    # here we override bind tracker so that we also define inactive time for this stimulator
    def bind_tracker(self, tracker):
        self._tracker = tracker

        roi_id = self._tracker._roi.idx
        try:
            channel = self._roi_to_channel[roi_id]
            self._inactivity_time_threshold_ms = round(channel ** 1.7) * 20 * 1000
        except KeyError:
            pass


class MiddleCrossingStimulator(BaseStimulator):
    _description = {"overview": "A stimulator to disturb animal as they cross the midline",
                    "arguments": [
                                    {"type": "number", "min": 0.0, "max": 1.0, "step":0.01, "name": "p", "description": "the probability to move the tube when a beam cross was detected","default":1.0},
                                    {"type": "date_range", "name": "date_range",
                                     "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                     "default": ""}
                                   ]}

    _HardwareInterfaceClass = SleepDepriverInterface
    _refractory_period = 60#s
    _roi_to_channel = {
            1:1,  3:2,  5:3,  7:4,  9:5,
            12:6, 14:7, 16:8, 18:9, 20:10
        }
    def __init__(self,
                 hardware_connection,
                 p=1.0,
                 date_range=""
                 ):
        """
        :param hardware_connection: the sleep depriver module hardware interface
        :type hardware_connection: :class:`~ethoscope.hardawre.interfaces.sleep_depriver_interface.SleepDepriverInterface`
        :param p: the probability of disturbing the animal when a beam cross happens
        :type p: float
        :return:
        """

        self._last_stimulus_time = 0
        self._p = p
        
        super(MiddleCrossingStimulator, self).__init__(hardware_connection,  date_range=date_range)

    def _decide(self):
        roi_id = self._tracker._roi.idx
        now = self._tracker.last_time_point
        if now - self._last_stimulus_time < self._refractory_period * 1000:
            return HasInteractedVariable(False), {}

        try:
            channel = self._roi_to_channel[roi_id]
        except KeyError:
            return HasInteractedVariable(False), {}

        positions = self._tracker.positions

        if len(positions) < 2:
            return HasInteractedVariable(False), {}

        if len(positions[-1]) != 1:
            raise Exception("This stimulator can only work with a single animal per ROI")

        roi_w = float(self._tracker._roi.longest_axis)
        x_t_zero = positions[-1][0]["x"] / roi_w - 0.5
        x_t_minus_one = positions[-2][0]["x"] / roi_w - 0.5

        # if roi_id == 12:
        #     print (roi_id, channel, roi_w, positions[-1][0]["x"], positions[-2][0]["x"], x_t_zero, x_t_minus_one)
        if (x_t_zero > 0) ^ (x_t_minus_one >0): # this is a change of sign

            if random.uniform(0,1) < self._p:
                self._last_stimulus_time = now
                return HasInteractedVariable(True), {"channel": channel}

        return HasInteractedVariable(False), {"channel": channel}





class OptomotorSleepDepriverSystematic(OptomotorSleepDepriver):
    _description = {"overview": "A stimulator to sleep deprive an animal using gear motors. See https://github.com/gilestrolab/ethoscope_hardware/tree/master/modules/gear_motor_sleep_depriver",
                    "arguments": [

                                    {"type": "number", "min": 1, "max": 3600*12, "step":1, "name": "interval", "description": "The recurence of the stimulus","default":120},
                                    {"type": "number", "min": 500, "max": 10000 , "step": 50, "name": "pulse_duration", "description": "For how long to deliver the stimulus(ms)", "default": 1000},
                                    {"type": "number", "min": 0, "max": 3, "step": 1, "name": "stimulus_type",  "description": "1 = opto, 2= moto", "default": 2},
                                    {"type": "date_range", "name": "date_range",
                                     "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                     "default": ""}
                                   ]}

    _HardwareInterfaceClass = OptoMotor
    _roi_to_channel_opto = {1:1, 3:3, 5:5, 7:7, 9:9,
                            12:23, 14:21,16:19, 18:17, 20:15}
    _roi_to_channel_moto = {1:0, 3:2, 5:4, 7:6, 9:8,
                            12:22, 14:20, 16:18, 18:16, 20:14}


    def __init__(self,
                 hardware_connection,
                 interval=120,  # s
                 pulse_duration = 1000,  #ms
                 stimulus_type = 2,  # 1 = opto, 2= moto, 3 = both
                 date_range=""
                 ):



        self._interval = interval  *1000 # ms used internally

        super(OptomotorSleepDepriverSystematic, self).__init__(hardware_connection, 0,0,
                                                               pulse_duration, stimulus_type,
                                                               date_range)

        self._t0 = 0

    def _decide(self):
        roi_id = self._tracker._roi.idx
        try:
            channel = self._roi_to_channel[roi_id]
        except KeyError:
            return HasInteractedVariable(False), {}
        now = self._tracker.last_time_point + roi_id *100
        if now - self._t0 > self._interval:
            dic = {"channel": channel}
            dic["duration"] = self._pulse_duration
            self._t0 = now
            return HasInteractedVariable(True), dic

        return HasInteractedVariable(False), {}



__author__ = 'quentin'

import numpy as np

class ScanException(BaseException):
    r"""
    A BaseException raised when `~_get_json` fails
    """
    pass


class EthoscopeException(Exception):

    def __init__(self, value, img=None):
        """
        A custom exception. It can store an image

        :param value: A value passed to the exception, generally a text message
        :param img: an image
        :type img: :class:`~numpy.ndarray`
        :return:
        """
        self.value = value
        if isinstance(img, np.ndarray):
            self.img = np.copy(img)
        else:
            self.img = None

    def __str__(self):
        return repr(self.value)

class IDOCException(EthoscopeException):
    r"""
    The same as EthoscopeException but with a name
    fitting for this package
    """
    pass


# # https://stackoverflow.com/a/4761058
# import sys
# sys.path.append('../src')
# import unittest
# from pathlib import Path
# import logging
# from src.interface.main import Interface


# class TestGeneric(unittest.TestCase):

#     def __init__(self, *args, **kwargs):

#         super(TestGeneric, self).__init__(*args, **kwargs)
#         self.store = None

#     def test_writable_store(self):
        
#         # Check dir exists
#         directory = Path(self.store).parents[0]
#         self.assertTrue(directory.is_dir())

#         # Check dir is writable
#         # https://stackoverflow.com/questions/2113427/determining-whether-a-directory-is-writeable
#         writable = False
#         try:
#             p = Path(directory, '.tmp.py')
#             p.touch()
#             p.unlink()
#             writable = True
#         except Exception:
#             logging.error("Trying to write the cache to a non writable dir")
        
#         self.assertTrue(writable)



import logging
import os
import os.path
import time
import traceback
import numpy as np
import cv2

from idoc.server.io.cameras import AdaptorCamera

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



class OpenCVCamera(AdaptorCamera):
    r"""
    Drive a Basler camera using pypylon.
    """

    def __init__(self, *args, timeout=5000, video_path=None, **kwargs):

        # TODO Put this stuff in the config
        self._max_failed_count = 10
        self._timeout = timeout
        self._video_path = video_path
        self._is_opened = False
        super().__init__(*args, **kwargs)

    def is_last_frame(self):
        return False

    def _time_stamp(self):
        now = time.time()
        return now - self._start_time

    def is_opened(self):
        return self._is_opened

    def _next_image(self):

        failed_count = 0
        # this while loop is designed to try several times,
        # not to be run continuosly
        # i.e. in normal conditions, it should be broken every time
        # the continously running loop is implemented BaseCamera.__iter__

        while failed_count < self._max_failed_count:

            ret, img = self.camera.read()
            if ret and isinstance(img, np.ndarray):
                break
            else:
                failed_count += 1
        
        # NOTE: This is not the most efficient way to guarantee BGR
        # And maybe I should make them gray from the start i.e. not support color
        # since they turn gray somewhere downstream
        if len(img.shape) == 2:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        self._validate(img)
        return img

    def open(self):

        try:
            super().open()
            self.camera = cv2.VideoCapture(self._video_path)
            self._is_opened = True
            ret, image = self.camera.read()
            logger.info("OpenCV camera loaded successfully")
            logger.info("Resolution of incoming frames: %dx%d", image.shape[1], image.shape[0])

        except Exception as error:
            logger.error(error)
            logger.error(traceback.print_exc())


    # called by BaseCamera.__exit__()
    def close(self):
        super().close()
        self.camera.release()
        self._is_opened = False


    def restart(self):
        r"""
        Attempt to restart a Basler camera
        """
        NotImplementedError

    @property
    def resolution(self):
        r"""
        Convenience function to return resolution of camera.
        Resolution = (number_horizontal_pixels, number_vertical_pixels)
        """
        width = self.camera.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = self.camera.get(cv2.CAP_PROP_FRAME_HEIGHT)
        
        self._settings["resolution"] = (width, height)
        return self._settings["resolution"]

    @resolution.setter
    def resolution(self, resolution):
        pass

    @property
    def shape(self):
        r"""
        Convenience function to return shape of camera
        Shape = (number_vertical_pixels, number_horizontal_pixels, number_channels)
        """
        width = self.camera.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = self.camera.get(cv2.CAP_PROP_FRAME_HEIGHT)

        return (width, height, 1)

    @property
    def framerate(self):
        return self.camera.get(cv2.CAP_PROP_FPS)



import warnings
import os.path
from idoc.configuration import IDOCConfiguration, MACHINE_NAME, MACHINE_ID

def main():
    # create the default config file
    IDOCConfiguration(warn_if_exists=True)
    
    if not os.path.exists(MACHINE_NAME):
        DEFAULT_NAME="IDOC_001"
        with open(MACHINE_NAME, "w") as filehandle:
            filehandle.write(f"{DEFAULT_NAME}\n")
            warnings.warn(
                f"Naming machine {DEFAULT_NAME}."\
                "If you wish to change the name,"\
                f"please update the contents of {MACHINE_NAME}")

    if not os.path.exists(MACHINE_ID):
        warnings.warn(
            f"File {MACHINE_ID} does not exist."\
            " Please create it and" \
            " enter a string of 32 alphanumeric (0-9 and a-f)"\
            " characters and a newline characters"
        )


import argparse
from idoc.server.core.control_thread import ControlThread

def list_options(category):
    """
    Return a list of str with the names of the classes that can be passed
    for a given category.
    """
    return [cls.__name__ for cls in ControlThread._option_dict[category]['possible_classes']]

def get_parser():
    # Arguments to follow the command, adding video, etc options
    parser = argparse.ArgumentParser(
        prog="IDOC - The Individual Drosophila Optogenetic Conditioner",
        description=(
            """
            A modular package to monitor and store the position of flies in separate chambers
            while running a preset paradigm of hardware events controlled by an Arduino board.
            A GUI is available by running the client.py script in the same computer.
            """
        ),
        epilog="Developed at the Liu Lab @ VIB-KU Leuven Center for Brain and Disease Research.",
        fromfile_prefix_chars='@', allow_abbrev=False
    )

    # General settings
    parser.add_argument("-D", "--debug", action='store_true', dest="debug")


    # Boolean flags
    ## Controller module
    parser.add_argument(
        "--control", action='store_true', dest='control',
        help="Turn on controller module"
    )
    parser.add_argument(
        "--no-control", action='store_false', dest='control',
        help="Turn off controller module"
    )

    ## Recognizer module
    parser.add_argument(
        "--recognize", action='store_true', dest='recognize',
        help="Turn on recognizer module"
    )

    parser.add_argument(
        "--no-recognize", action='store_false', dest='recognize',
        help="Turn off recognizer module"
    )
    parser.add_argument(
        "--wrap", action='store_true', dest="wrap",
        help="If pasased a video, play it in a loop. Useful for debugging"
    )
    parser.add_argument(
        "--no-wrap", action='store_false', dest="wrap",
        help="If pasased a video, DO NOT play it in a loop. Useful for analysis"
    )

    parser.add_argument(
        "--use-wall-clock", action='store_true', dest='use_wall_clock',
        help="""
        Default. If passed, the time is computed using the output of time.time()
        """
    )
    parser.add_argument(
        "--no-use-wall-clock", action='store_false', dest='use_wall_clock',
        help="""
        If passed, the time is computed using the CAP_PROP_POS_MSEC property
        of the OpenCV capture class. This is only compatible with the
        OpenCV camera and when analyzing a video offline.
        """
    )

    # # Module settings
    parser.add_argument(
        "-m", "--mapping-path", type=str, dest="mapping_path",
        help="Absolute path to csv providing pin number-pin name mapping"
    )
    parser.add_argument(
        "--paradigm-path", type=str, dest="paradigm_path",
        help="Absolute path to csv providing the Arduino top level paradigm"
    )

    parser.add_argument(
        "--arduino_port", type=str,
        help="Absolute path to the Arduino port. Usually '/dev/ttyACM0' in Linux and COM in Windows"
    )

    parser.add_argument(
        "-f", "--framerate", type=int,
        help="Frames per second in the opened stream, overrides config."
    )
    parser.add_argument(
        "-v", "--video-path", type=str, dest='video_path',
        help=
        """
        If offline tracking is desired, location of the input video file.
        Only possible if OpenCVCamera class is passed in --camera.
        """
    )


    parser.add_argument(
        "-a", "--adaptation-time", type=int, dest="adaptation_time",
        help=
        """
        Number of seconds to wait before starting paradigm and saving tracking data
        from the moment this script is started. If not provided, it is taken from config.
        """
    )

    parser.add_argument(
        "-d", "--duration", type=int, dest="max_duration",
        help=
        """
        Maximum number of seconds the software will run. If not provided, it is taken from config.
        """
    )

    parser.add_argument(
        "--run", action='store_true', dest='run',
        help="Batch run"
    )
    parser.add_argument(
        "--no-run", action='store_false', dest='run',
        help="Batch run"
    )

    parser.add_argument(
        "--start-datetime", type=str, dest='start_datetime',
        help=""
    )

    # User input for classes
    parser.add_argument("--board", type=str, choices=list_options("board"))
    parser.add_argument("--camera", type=str, help="Stream source", choices=list_options("camera"))
    parser.add_argument("--drawer", type=str, choices=list_options("drawer"))
    parser.add_argument("--roi-builder", type=str, dest="roi_builder", choices=list_options("roi_builder"))
    parser.add_argument("--result-writer", type=str, dest="result_writer", choices=list_options("result_writer"))
    parser.add_argument("--tracker", type=str, choices=list_options("tracker"))


    # General application module
    parser.add_argument("-e", "--experimenter", type=str, help="Add name of the experimenter/operator")

    # Bottle settings
    parser.add_argument("-p", "--port", type=int)


    parser.set_defaults(
        control=False, recognize=False,
        wrap=None,
        default=True,
        max_duration=None
    )

    return parser



import logging
from threading import Thread

from idoc.server.core.control_thread import ControlThread
from idoc.helpers import get_server
from idoc.configuration import IDOCConfiguration
from idoc.server.core.control_thread import ControlThread
from .parser import get_parser, list_options
from .server import run, stop

logger = logging.getLogger(__name__)

config = IDOCConfiguration()
ARGS = vars(get_parser().parse_args())
from idoc import __version__ as VERSION
from idoc.helpers import get_machine_id

machine_id = get_machine_id()
RESULT_DIR = config.content["folders"]["results"]["path"]
PORT = config.content["network"]["port"]
DEBUG = config.content["network"]["port"]


control = ControlThread(
    machine_id=machine_id,
    version=VERSION,
    result_dir=RESULT_DIR,
    user_data=ARGS,
)


def offline():

    ARGS["control"] = False
    ARGS["recognize"] = True

    control = ControlThread(
        machine_id=machine_id,
        version=VERSION,
        result_dir=RESULT_DIR,
        user_data=ARGS,
    )

    server_thread = Thread(target=run, name="bottle", args=(PORT,))

    control.start()
    server_thread.start()
    control.join()
    stop()